from reinforced_lib import RLib
from reinforced_lib.agents.mab import EGreedy


if __name__ == '__main__':
    rl = RLib(
        agent_type=EGreedy,
        agent_params={'n_arms': 4, 'e': 0.1},
        no_ext_mode=True
    )

    print(rl.observation_space)

    observations = {
        'action': 3,
        'reward': 1.0
    }

    action = rl.sample(update_observations=observations)
    print(action)

    observations = {
        'action': 3,
        'reward': 1.0
    }

    action = rl.sample(update_observations=observations)
    print(action)



import os
import unittest
from glob import glob

import jax.numpy as jnp
import optax
from flax import linen as nn

from reinforced_lib.agents.deep import DDPG
from reinforced_lib.agents.mab import ThompsonSampling
from reinforced_lib.exts import Gymnasium
from reinforced_lib.rlib import RLib


class TestRLibToTflite(unittest.TestCase):
    def test_sample_only_export(self):
        try:
            import tensorflow as tf
        except ModuleNotFoundError:
            return

        rl = RLib(
            agent_type=ThompsonSampling,
            agent_params={'n_arms': 4},
            no_ext_mode=True
        )
        rl.to_tflite(agent_id=0, sample_only=True)

        with open(glob(f'{os.path.expanduser("~")}/rlib-ThompsonSampling-0-*-init.tflite')[0], 'rb') as f:
            interpreter = tf.lite.Interpreter(model_content=f.read())
            interpreter.allocate_tensors()
            interpreter.invoke()
            key = [interpreter.get_tensor(od['index']) for od in interpreter.get_output_details()]

        with open(glob(f'{os.path.expanduser("~")}/rlib-ThompsonSampling-0-*-sample.tflite')[0], 'rb') as f:
            interpreter = tf.lite.Interpreter(model_content=f.read())
            interpreter.allocate_tensors()
            input_details = interpreter.get_input_details()
            ins = key + [jnp.arange(4).astype(float)]

            for d, a in zip(input_details, ins):
                interpreter.set_tensor(d['index'], a)

            interpreter.invoke()
            action, key = [interpreter.get_tensor(od['index']) for od in interpreter.get_output_details()]

        assert isinstance(action, (int, int))
        assert isinstance(key, jnp.ndarray)
        assert key.shape == (2,)

    def test_full_export(self):
        try:
            import tensorflow as tf
        except ModuleNotFoundError:
            return

        rl = RLib(
            agent_type=ThompsonSampling,
            agent_params={'n_arms': 4},
            no_ext_mode=True
        )
        rl.to_tflite()

        with open(glob(f'{os.path.expanduser("~")}/rlib-ThompsonSampling-*-init.tflite')[0], 'rb') as f:
            interpreter = tf.lite.Interpreter(model_content=f.read())
            interpreter.allocate_tensors()
            interpreter.invoke()
            state = [interpreter.get_tensor(od['index']) for od in interpreter.get_output_details()]

        with open(glob(f'{os.path.expanduser("~")}/rlib-ThompsonSampling-*-update.tflite')[0], 'rb') as f:
            interpreter = tf.lite.Interpreter(model_content=f.read())
            interpreter.allocate_tensors()
            input_details = interpreter.get_input_details()
            ins = state + [
                jnp.asarray(2, dtype=int),
                jnp.asarray([1.], dtype=float),
                jnp.asarray([1], dtype=int),
                jnp.asarray([0], dtype=int)
            ]

            for d, a in zip(input_details, ins):
                interpreter.set_tensor(d['index'], a)

            interpreter.invoke()
            state = [interpreter.get_tensor(od['index']) for od in interpreter.get_output_details()]

        with open(glob(f'{os.path.expanduser("~")}/rlib-ThompsonSampling-*-sample.tflite')[0], 'rb') as f:
            interpreter = tf.lite.Interpreter(model_content=f.read())
            interpreter.allocate_tensors()
            input_details = interpreter.get_input_details()
            ins = state + [jnp.arange(4).astype(float)]

            for d, a in zip(input_details, ins):
                interpreter.set_tensor(d['index'], a)

            interpreter.invoke()
            action, key, state_alpha, state_beta = [interpreter.get_tensor(od['index']) for od in interpreter.get_output_details()]

        assert isinstance(action, (int, int))
        assert isinstance(key, jnp.ndarray)
        assert key.shape == (2,)
        assert isinstance(state_alpha, jnp.ndarray)
        assert state_alpha.shape == (4, 1)
        assert isinstance(state_beta, jnp.ndarray)
        assert state_beta.shape == (4, 1)

    def test_drl_sample_only_export(self):
        try:
            import tensorflow as tf
        except ModuleNotFoundError:
            return

        class QNetwork(nn.Module):
            @nn.compact
            def __call__(self, s, a):
                s = nn.Dense(64)(s)
                s = nn.relu(s)
                return nn.Dense(1)(s)

        class ANetwork(nn.Module):
            @nn.compact
            def __call__(self, s):
                s = nn.Dense(32)(s)
                s = nn.relu(s)
                s = nn.Dense(32)(s)
                s = nn.relu(s)
                return nn.Dense(1)(s)

        rl = RLib(
            agent_type=DDPG,
            agent_params={
                'q_network': QNetwork(),
                'a_network': ANetwork(),
                'q_optimizer': optax.adam(2e-3),
                'a_optimizer': optax.adam(1e-3)
            },
            ext_type=Gymnasium,
            ext_params={'env_id': 'Pendulum-v1'}
        )
        rl.to_tflite(agent_id=0, sample_only=True)

    def test_drl_full_export(self):
        try:
            import tensorflow as tf
        except ModuleNotFoundError:
            return

        class QNetwork(nn.Module):
            @nn.compact
            def __call__(self, s, a):
                s = nn.Dense(64)(s)
                s = nn.relu(s)
                return nn.Dense(1)(s)

        class ANetwork(nn.Module):
            @nn.compact
            def __call__(self, s):
                s = nn.Dense(32)(s)
                s = nn.relu(s)
                s = nn.Dense(32)(s)
                s = nn.relu(s)
                return nn.Dense(1)(s)

        rl = RLib(
            agent_type=DDPG,
            agent_params={
                'q_network': QNetwork(),
                'a_network': ANetwork(),
                'q_optimizer': optax.adam(2e-3),
                'a_optimizer': optax.adam(1e-3)
            },
            ext_type=Gymnasium,
            ext_params={'env_id': 'Pendulum-v1'}
        )
        rl.to_tflite()


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
from typing import Callable

from reinforced_lib.agents import BaseAgent
from reinforced_lib.logs import BaseLogger, Source, SourceType
from reinforced_lib.utils.exceptions import IncorrectLoggerTypeError, IncorrectSourceTypeError
from reinforced_lib.utils import is_scalar, is_array, is_dict


class LogsObserver:
    """
    Class responsible for managing singleton instances of the loggers, initialization and finalization
    of the loggers, and passing the logged values to the appropriate loggers and their methods.
    """

    def __init__(self) -> None:
        self._logger_instances = {}
        self._logger_sources = defaultdict(list)

        self._observations_loggers = defaultdict(list)
        self._agent_state_loggers = defaultdict(list)
        self._metrics_loggers = defaultdict(list)
        self._custom_loggers = defaultdict(list)

    def add_logger(self, source: Source, logger_type: type, logger_params: dict[str, any]) -> None:
        """
        Initializes a singleton instance of the logger and connects a given source with that logger.

        Parameters
        ----------
        source : Source
            Source to connect.
        logger_type : type
            Type of the selected loger.
        logger_params : dict
            Parameters of the selected logger.
        """

        if not issubclass(logger_type, BaseLogger):
            raise IncorrectLoggerTypeError(logger_type)

        if isinstance(source, tuple):
            if len(source) != 2 or not isinstance(source[0], str) or not hasattr(source[1], 'name'):
                raise IncorrectSourceTypeError(type(source))
        elif source is not None and not isinstance(source, str):
            raise IncorrectSourceTypeError(type(source))

        logger = self._logger_instances.get(logger_type, logger_type(**logger_params))

        if isinstance(source, tuple):
            if source[1] == SourceType.OBSERVATION:
                self._observations_loggers[logger].append((source, source[0]))
            elif source[1] == SourceType.STATE:
                self._agent_state_loggers[logger].append((source, source[0]))
            elif source[1] == SourceType.METRIC:
                self._metrics_loggers[logger].append((source, source[0]))
        elif isinstance(source, str):
            self._observations_loggers[logger].append((source, source))
            self._agent_state_loggers[logger].append((source, source))
            self._metrics_loggers[logger].append((source, source))

        self._custom_loggers[logger] = [(None, None)]
        self._logger_sources[logger].append(source)
        self._logger_instances[logger_type] = logger

    def init_loggers(self):
        """
        Initializes all loggers by calling their ``init`` method.
        """

        for logger, sources in self._logger_sources.items():
            logger.init(sources)

    def finish_loggers(self):
        """
        Finalizes the work of all loggers by calling their ``finish`` method.
        """

        for logger in self._logger_sources.keys():
            logger.finish()

    def update_observations(self, observations: any) -> None:
        """
        Passes new observations to the loggers.

        Parameters
        ----------
        observations : dicy or any
            Observations received by the agent.
        """

        if isinstance(observations, dict):
            self._update(self._observations_loggers, lambda name: observations.get(name, None))
        else:
            self._update(self._observations_loggers, lambda name: observations)

    def update_agent_state(self, agent_state: BaseAgent) -> None:
        """
        Passes the agent state to the loggers.

        Parameters
        ----------
        agent_state : BaseAgent
            Current agent state.
        """

        self._update(self._agent_state_loggers, lambda name: getattr(agent_state, name, None))

    def update_metrics(self, metric: any, metric_name: str) -> None:
        """
        Passes metrics to loggers.

        Parameters
        ----------
        metric : any
            Metric value.
        metric_name : str
            Name of the metric.
        """

        self._update(self._metrics_loggers, lambda name: metric if name == metric_name else None)

    def update_custom(self, value: any, name: str) -> None:
        """
        Passes values provided by the user to the loggers.

        Parameters
        ----------
        value : any
            Value to log.
        name : str
            Name of the value.
        """

        self._update(self._custom_loggers, lambda _: (name, value))

    @staticmethod
    def _update(loggers: dict[BaseLogger, list[Source]], get_value: Callable) -> None:
        """
        Passes values to the appropriate loggers and method based on the type and the source of the value.

        Parameters
        ----------
        loggers : dict
            Dictionary with the logger instances and the connected sources.
        get_value : callable
            Function that gets the selected value from the observations, state, or metrics.
        """

        for logger, sources in loggers.items():
            for source, name in sources:
                if (value := get_value(name)) is not None:
                    if (custom := name is None):
                        source, value = value

                    if is_scalar(value):
                        logger.log_scalar(source, value, custom)
                    elif is_dict(value):
                        logger.log_dict(source, value, custom)
                    elif is_array(value):
                        logger.log_array(source, value, custom)
                    else:
                        logger.log_other(source, value, custom)



import gymnasium as gym


class NoAgentError(Exception):
    """
    Raised when no agent is specified.
    """

    def __str__(self) -> str:
        return 'No agent is specified.'


class NoExtensionError(Exception):
    """
    Raised when no extension is specified.
    """

    def __str__(self) -> str:
        return 'No extension is specified.'


class IncorrectTypeError(Exception):
    """
    Raised when provided class type is incorrect.

    Parameters
    ----------
    provided_type : type, optional
        Type provided by the user.
    expected_module : str, optional
        Name of the module that ``provided_type`` should match.
    """

    def __init__(self, provided_type: type = None, expected_module: str = None) -> None:
        self._provided_type = provided_type.__name__ if provided_type else 'Provided type'
        self._expected_module = expected_module if expected_module else ''

    def __str__(self) -> str:
        return f'{self._provided_type} is not a valid {self._expected_module} type.'


class IncorrectAgentTypeError(IncorrectTypeError):
    """
    Raised when provided agent does not inherit from the BaseAgent class.

    Parameters
    ----------
    provided_type : type
        Type provided by the user.
    """

    def __init__(self, provided_type: type) -> None:
        super().__init__(provided_type, 'agent')


class IncorrectExtensionTypeError(IncorrectTypeError):
    """
    Raised when provided extension does not inherit from the BaseExt class.

    Parameters
    ----------
    provided_type : type
        Type provided by the user.
    """

    def __init__(self, provided_type: type) -> None:
        super().__init__(provided_type, 'extension')


class IncorrectLoggerTypeError(IncorrectTypeError):
    """
    Raised when provided logger does not inherit from the BaseLogger class.

    Parameters
    ----------
    provided_type : type
        Type provided by the user.
    """

    def __init__(self, provided_type: type) -> None:
        super().__init__(provided_type, 'logger')


class ForbiddenOperationError(Exception):
    """
    Raised when the user performs a forbidden operation.
    """

    def __str__(self) -> str:
        return 'Forbidden operation.'


class ForbiddenAgentChangeError(ForbiddenOperationError):
    """
    Raised when the user changes the agent type after the first agent instance has been initialized.
    """

    def __str__(self) -> str:
        return 'Cannot change agent type after the first agent instance is initialized.'


class ForbiddenExtensionChangeError(ForbiddenOperationError):
    """
    Raised when the user changes the extension type after the first agent instance has been initialized.
    """

    def __str__(self) -> str:
        return 'Cannot change extension type after the first agent instance is initialized.'


class ForbiddenExtensionSetError(ForbiddenOperationError):
    """
    Raised when the user sets the extension type when ``no_ext_mode`` is enabled.
    """

    def __str__(self) -> str:
        return 'Cannot set extension type when \'no_ext_mode\' is enabled.'


class ForbiddenLoggerSetError(ForbiddenOperationError):
    """
    Raised when the user adds a new logger after the first step has been made.
    """

    def __str__(self) -> str:
        return 'Cannot add new loggers type after the first step has been made.'


class IncorrectSpaceError(Exception):
    """
    Raised when an unknown space is provided, for example a custom Gymnasium space.
    """

    def __str__(self) -> str:
        return 'Cannot find corresponding Gymnasium space.'


class UnimplementedSpaceError(Exception):
    """
    Raised when an observation space is required but not implemented.
    """

    def __str__(self) -> str:
        return 'Appropriate observation space is not implemented.'


class IncompatibleSpacesError(Exception):
    """
    Raised when the observation spaces of two different modules are not compatible.

    Parameters
    ----------
    ext_space : gym.spaces.Space
        Observation space of the extension.
    agent_space : gym.spaces.Space
        Observation space of the agent.
    """

    def __init__(self, ext_space: gym.spaces.Space, agent_space: gym.spaces.Space) -> None:
        self._ext_space = ext_space
        self._agent_space = agent_space

    def __str__(self) -> str:
        return f'Agents space of type {self._agent_space} is not compatible ' \
               f'with extension space of type {self._ext_space}.'


class NoDefaultParameterError(Exception):
    """
    Raised when the extension does not define a default parameter value for the agent.

    Parameters
    ----------
    extension_type : type
        Type of the used extension.
    parameter_name : str
        Name of the missing parameter.
    parameter_type : gym.spaces.Space
        Type of the missing parameter.
    """

    def __init__(self, extension_type: type, parameter_name: str, parameter_type: gym.spaces.Space) -> None:
        self._extension_name = extension_type.__name__
        self._parameter_name = parameter_name
        self._parameter_type = parameter_type

    def __str__(self) -> str:
        return f'Extension {self._extension_name} does not provide parameter ' \
               f'{self._parameter_name} of type {self._parameter_type}.'


class UnsupportedLogTypeError(Exception):
    """
    Raised when the user logs values that are not supported by the logger.

    Parameters
    ----------
    logger_type : type
        Type of the used logger.
    log_type : type
        Type of the logged value.
    """

    def __init__(self, logger_type: type, log_type: type) -> None:
        self._logger_name = logger_type.__name__
        self._log_name = log_type.__name__

    def __str__(self) -> str:
        return f'Logger {self._logger_name} does not support logging {self._log_name}.'


class IncorrectSourceTypeError(IncorrectTypeError):
    """
    Raised when the provided source is not a correct source type (i.e., ``Union[Tuple[str, SourceType], str]``).

    Parameters
    ----------
    provided_type : type
        Type provided by the user.
    """

    def __init__(self, provided_type: type) -> None:
        super().__init__(provided_type, 'source')


class UnsupportedCustomLogsError(Exception):
    """
    Raised when the user tries to log custom values with a logger that does not support custom logging.

    Parameters
    ----------
    logger_type : type
        Type of the used logger.
    """

    def __init__(self, logger_type: type) -> None:
        self._logger_name = logger_type.__name__

    def __str__(self) -> str:
        return f'Logger {self._logger_name} does not support custom logging.'



import os.path
from collections import defaultdict
from copy import deepcopy

import jax.numpy as jnp
import matplotlib.pyplot as plt
from chex import Array, Scalar, Numeric

from reinforced_lib.logs import BaseLogger, Source
from reinforced_lib.utils import timestamp


class PlotsLogger(BaseLogger):
    r"""
    Logger that presents and saves values as matplotlib plots. Offers smoothing of the curve, scatter plots, and
    multiple curves in a single chart (while logging arrays). ``PlotsLogger`` is able to synchronize the logged
    values in time. This means that if the same source is logged less often than other sources, the step will be
    increased accordingly to maintain the appropriate spacing between the values on the x-axis.

    Parameters
    ----------
    plots_dir : str, default="~"
        Output directory for the plots.
    plots_ext : str, default="svg"
        Extension of the saved plots.
    plots_smoothing : float, default=0.6
        Weight of the exponential moving average (EMA/EWMA) [1]_ used for smoothing. :math:`\alpha \in [0, 1)`.
    plots_scatter : bool, default=False
        Set to ``True`` if you want to generate a scatter plot instead of a line plot.
        ``plots_smoothing`` parameter does not apply to the scatter plots.
    plots_sync_steps : bool, default=False
        Set to ``True`` if you want to synchronize the logged values in time.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average
    """

    def __init__(
            self,
            plots_dir: str = None,
            plots_ext: str = 'pdf',
            plots_smoothing: Scalar = 0.6,
            plots_scatter: bool = False,
            plots_sync_steps: bool = False,
            **kwargs
    ) -> None:
        assert 1 > plots_smoothing >= 0

        super().__init__(**kwargs)

        self._dir = plots_dir if plots_dir else os.path.expanduser("~")
        self._ext = plots_ext
        self._smoothing = plots_smoothing
        self._scatter = plots_scatter
        self._sync_steps = plots_sync_steps

        self._current_values = set()
        self._step = 0

        self._values = defaultdict(list)
        self._steps = defaultdict(list)

    def finish(self) -> None:
        """
        Shows the generated plots and saves them to the output directory with the specified extension
        (the names of the files follow the pattern ``"rlib-plot-[source]-[date]-[time].[ext]"``).
        """

        def exponential_moving_average(values: list, weight: Scalar) -> list:
            smoothed = [values[0]]

            for value in values[1:]:
                smoothed.append((1 - weight) * value + weight * smoothed[-1])

            return smoothed

        def lineplot(values: list, steps: list, alpha: Scalar = 1.0, label: bool = False) -> None:
            values = jnp.array(values).squeeze()

            if values.ndim == 1:
                plt.plot(steps, values, alpha=alpha, c='C0')
            elif values.ndim == 2:
                for i, val in enumerate(jnp.array(values).T):
                    plt.plot(steps, val, alpha=alpha, c=f'C{i % 10}', label=i if label else '')
                plt.legend()
        
        def scatterplot(values: list, steps: list, label: bool = False) -> None:
            values = jnp.array(values).squeeze()

            if values.ndim == 1:
                plt.scatter(steps, values, c='C0', marker='.', s=4)
            elif values.ndim == 2:
                for i, val in enumerate(jnp.array(values).T):
                    plt.scatter(steps, val, c=f'C{i % 10}', label=i if label else '', marker='.', s=4)
                plt.legend()

        for name, values in self._values.items():
            filename = f'rlib-plot-{name}-{timestamp()}.{self._ext}'

            if self._scatter:
                scatterplot(values, self._steps[name], True)
            else:
                smoothed = exponential_moving_average(values, self._smoothing)
                lineplot(values, self._steps[name], alpha=0.3)
                lineplot(smoothed, self._steps[name], label=True)
            
            plt.title(name)
            plt.xlabel('step')
            plt.savefig(os.path.join(self._dir, filename), bbox_inches='tight', dpi=300)
            plt.show()

    def log_scalar(self, source: Source, value: Scalar, *_) -> None:
        """
        Adds a given scalar to the plot values.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        """

        self._log(source, value)

    def log_array(self, source: Source, value: Array, *_) -> None:
        """
        Log values from an array to the same plot. Creates multiple line plots for each value in the array.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        """

        self._log(source, value)

    def _log(self, source: Source, value: Numeric) -> None:
        """
        Adds a given value to the plot.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Numeric
            Value to log.
        """

        name = self.source_to_name(source)

        if self._sync_steps:
            if name in self._current_values:
                self._step += 1
                self._current_values.clear()

            self._current_values.add(name)
            step = self._step
        else:
            step = self._steps[name][-1] + 1 if self._steps[name] else 0

        self._values[name].append(deepcopy(value))
        self._steps[name].append(step)



import os
import pickle
from copy import deepcopy
from itertools import product
from typing import Union

import cloudpickle
import gymnasium as gym
import jax.random
import lz4.frame
from chex import dataclass

from reinforced_lib.agents import BaseAgent
from reinforced_lib.exts import BaseExt
from reinforced_lib.logs import Source
from reinforced_lib.logs.logs_observer import LogsObserver
from reinforced_lib.utils.exceptions import *
from reinforced_lib.utils import is_scalar, timestamp


@dataclass
class AgentContainer:
    """
    Class containing the state of a given agent and all its dependencies.

    Attributes
    ----------
    state : BaseAgent
        Current state of the agent.
    key : PRNGKey
        A PRNG key used as the random key.
    action : any
        Action selected by the agent.
    step : int
        Current step of the agent.
    """

    state: BaseAgent
    key: jax.random.PRNGKey
    action: any
    step: int


class RLib:
    """
    Main class of the library. Exposes a simple and intuitive interface to use the library.

    Parameters
    ----------
    agent_type : type, optional
        Type of the selected agent. Must inherit from the ``BaseAgent`` class.
    agent_params : dict, optional
        Parameters of the selected agent.
    ext_type : type, optional
        Type of the selected extension. Must inherit from the ``BaseExt`` class.
    ext_params : dict, optional
        Parameters of the selected extension.
    logger_types : type or list[type], optional
        Types of the selected loggers. Must inherit from the ``BaseLogger`` class.
    logger_sources : Source or list[Source], optional
        Sources to log.
    logger_params : dict, optional
        Parameters of the selected loggers.
    no_ext_mode : bool, default=False
        Pass observations directly to the agent (do not use the extensions).
    auto_checkpoint : int, optional
        Automatically save the experiment every ``auto_checkpoint`` steps.
        If ``None``, the automatic checkpointing is disabled.
    auto_checkpoint_path : str, optional, default=~
        Path to the directory where the automatic checkpoints will be saved.
    """

    def __init__(
            self, *,
            agent_type: type = None,
            agent_params: dict[str, any] = None,
            ext_type: type = None,
            ext_params: dict[str, any] = None,
            logger_types: Union[type, list[type]] = None,
            logger_sources: Union[Source, list[Source]] = None,
            logger_params: dict[str, any] = None,
            no_ext_mode: bool = False,
            auto_checkpoint: int = None,
            auto_checkpoint_path: str = None
    ) -> None:
        self._lz4_ext = ".pkl.lz4"
        self._default_path = os.path.expanduser("~")
        self._auto_checkpoint = auto_checkpoint
        self._auto_checkpoint_path = auto_checkpoint_path if auto_checkpoint_path else self._default_path

        self._agent = None
        self._agent_type = agent_type
        self._agent_params = deepcopy(agent_params)
        self._agent_containers = []

        self._ext = None
        self._no_ext_mode = no_ext_mode
        self._ext_type = ext_type
        self._ext_params = deepcopy(ext_params)

        self._logger_types = deepcopy(logger_types)
        self._logger_sources = deepcopy(logger_sources)
        self._logger_params = deepcopy(logger_params)
        self._logs_observer = LogsObserver()
        self._init_loggers = True
        self._cumulative_reward = 0.0

        if ext_type:
            self.set_ext(self._ext_type, self._ext_params)

        if agent_type:
            self.set_agent(self._agent_type, self._agent_params)

        if logger_types:
            self.set_loggers(self._logger_types, self._logger_sources, self._logger_params)

    def __del__(self) -> None:
        """
        Automatically finalizes the library work.
        """

        self.finish()

    def finish(self) -> None:
        """
        Used to explicitly finalize the library's work. In particular, it finishes the logger's work.
        """

        self._logs_observer.finish_loggers()

    def set_agent(self, agent_type: type, agent_params: dict = None) -> None:
        """
        Initializes an agent of type ``agent_type`` with parameters ``agent_params``. The agent type must inherit from
        the ``BaseAgent`` class. The agent type cannot be changed after the first agent instance has been initialized.

        Parameters
        ----------
        agent_type : type
            Type of the selected agent. Must inherit from the ``BaseAgent`` class.
        agent_params : dict, optional
            Parameters of the selected agent.
        """

        if len(self._agent_containers) > 0:
            raise ForbiddenAgentChangeError()

        if not issubclass(agent_type, BaseAgent):
            raise IncorrectAgentTypeError(agent_type)

        self._agent_type = agent_type
        self._agent_params = deepcopy(agent_params)

        if not self._no_ext_mode and self._ext:
            agent_params = self._ext.get_agent_params(agent_type, agent_type.parameter_space(), self._agent_params)
            self._agent = agent_type(**agent_params)
            self._ext.setup_transformations(self._agent.update_observation_space, self._agent.sample_observation_space)
        else:
            agent_params = self._agent_params if self._agent_params else {}
            self._agent = agent_type(**agent_params)

    def set_ext(self, ext_type: type, ext_params: dict = None) -> None:
        """
        Initializes an extension of type ``ext_type`` with parameters ``ext_params``. The extension type must inherit
        from the ``BaseExt`` class. The extension type cannot be changed after the first agent instance has been
        initialized.

        Parameters
        ----------
        ext_type : type
            Type of selected extension. Must inherit from the ``BaseExt`` class.
        ext_params : dict, optional
            Parameters of the selected extension.
        """

        if self._no_ext_mode:
            raise ForbiddenExtensionSetError()

        if len(self._agent_containers) > 0:
            raise ForbiddenExtensionChangeError()

        if not issubclass(ext_type, BaseExt):
            raise IncorrectExtensionTypeError(ext_type)

        self._ext_type = ext_type
        self._ext_params = deepcopy(ext_params)

        ext_params = self._ext_params if self._ext_params else {}
        self._ext = ext_type(**ext_params)

        if self._agent:
            agent_params = self._ext.get_agent_params(
                self._agent_type,
                self._agent_type.parameter_space(),
                self._agent_params
            )
            self._agent = self._agent_type(**agent_params)
            self._ext.setup_transformations(self._agent.update_observation_space, self._agent.sample_observation_space)

    def set_loggers(
            self,
            logger_types: Union[type, list[type]],
            logger_sources: Union[Source, list[Source]] = None,
            logger_params: dict[str, any] = None
    ) -> None:
        r"""
        Initializes loggers of types ``logger_types`` with parameters ``logger_params``. The logger types must inherit
        from the ``BaseLogger`` class. The logger types cannot be changed after the first agent instance has been
        initialized. ``logger_types`` and ``logger_sources`` can be objects or lists of objects, the function broadcasts
        them so that all loggers are connected to all sources. The ``logger_sources`` parameter specifies the sources
        to log. A source can be a name (e.g., "action") or tuple containing the name and the ``SourceType`` (e.g.,
        ``("action", SourceType.OBSERVATION)``). If the name itself is inconclusive (e.g., it occurs as a metric and
        as an observation), the behavior depends on the implementation of the logger.

        Parameters
        ----------
        logger_types : type or list[type]
            Types of the selected loggers.
        logger_sources : Source or list[Source], optional
            Sources to log.
        logger_params : dict, optional
            Parameters of the selected loggers.
        """

        if not self._init_loggers:
            raise ForbiddenLoggerSetError()

        self._logger_types = deepcopy(logger_types)
        self._logger_sources = deepcopy(logger_sources)
        self._logger_params = deepcopy(logger_params)

        logger_params =  self._logger_params if  self._logger_params else {}
        logger_types, logger_sources = self._object_to_list(self._logger_types), self._object_to_list(self._logger_sources)

        for logger_type, source in product(logger_types, logger_sources):
            self._logs_observer.add_logger(source, logger_type, logger_params)

    @staticmethod
    def _object_to_list(obj: Union[any, list[any]]) -> list[any]:
        return obj if isinstance(obj, list) else [obj]

    @property
    def observation_space(self) -> gym.spaces.Space:
        """
        Returns the observation space of the selected extension (or agent, if ``no_ext_mode`` is set).

        Returns
        -------
        gym.spaces.Space
            Observation space of the selected extension or agent.
        """

        if self._no_ext_mode:
            if not self._agent:
                raise NoAgentError()
            else:
                return gym.spaces.Dict({
                    'update_observation_space': self._agent.update_observation_space,
                    'sample_observation_space': self._agent.sample_observation_space
                })
        else:
            if not self._ext:
                raise NoExtensionError()
            else:
                return self._ext.observation_space

    @property
    def action_space(self) -> gym.spaces.Space:
        """
        Returns the action space of the selected agent.

        Returns
        -------
        gym.spaces.Space
            Action space of the selected agent.
        """

        if not self._agent:
            raise NoAgentError()

        return self._agent.action_space

    def init(self, seed: int = 42) -> int:
        """
        Initializes a new instance of the agent.

        Parameters
        ----------
        seed : int, default=42
            Number used to initialize the JAX pseudo-random number generator.

        Returns
        -------
        int
            Identifier of the created instance.
        """

        agent_id = len(self._agent_containers)
        init_key, key = jax.random.split(jax.random.key(seed))

        self._agent_containers.append(AgentContainer(
            state=self._agent.init(init_key),
            key=key,
            action=None,
            step=0
        ))

        return agent_id

    def sample(
            self,
            *args,
            agent_id: int = 0,
            is_training: bool = True,
            update_observations: Union[dict, tuple, any] = None,
            sample_observations: Union[dict, tuple, any] = None,
            **kwargs
    ) -> any:
        """
        Takes the extension state as an input, updates the agent state, and returns the next action selected by
        the agent. If ``no_ext_mode`` is disabled, observations are passed by args and kwargs (the observations must
        match the extension observation space). If ``no_ext_mode`` is enabled, observations must be passed
        by the ``update_observations`` and ``sample_observations`` parameters (the observations must match the agent's
        ``update_observation_space`` and ``sample_observation_space``). If there are no agent instances initialized,
        the method automatically initializes the first instance. If the ``is_training`` flag is set, the ``update`` and
        ``sample`` agent methods will be called. Otherwise, only the ``sample`` method will be called.

        Parameters
        ----------
        *args : tuple
            Environment observations.
        agent_id : int, default=0
            The identifier of the agent instance.
        is_training : bool
            Flag indicating whether the agent state should be updated in this step.
        update_observations : dict or tuple or any, optional
            Observations used when ``no_ext_mode`` is enabled (must match agent's ``update_observation_space``).
        sample_observations : dict or tuple or any, optional
            Observations used when ``no_ext_mode`` is enabled (must match agent's ``sample_observation_space``).
        **kwargs : dict
            Environment observations.

        Returns
        -------
        any
            Action selected by the agent.
        """

        if not self._agent:
            raise NoAgentError()

        if not self._no_ext_mode and not self._ext:
            raise NoExtensionError()

        update_observations = update_observations if update_observations else {}
        sample_observations = sample_observations if sample_observations else {}

        if self._init_loggers:
            self._logs_observer.init_loggers()
            self._init_loggers = False

        if len(self._agent_containers) == 0:
            self.init()

        key, update_key, sample_key = jax.random.split(self._agent_containers[agent_id].key, 3)
        state = self._agent_containers[agent_id].state
        action = self._agent_containers[agent_id].action
        step = self._agent_containers[agent_id].step

        if not self._no_ext_mode:
            update_observations, sample_observations = self._ext.transform(*args, action=action, **kwargs)

        all_observations = kwargs
        if isinstance(update_observations, dict) and isinstance(sample_observations, dict):
            all_observations |= update_observations
            all_observations |= sample_observations
            self._logs_observer.update_observations(all_observations)
        else:
            self._logs_observer.update_observations(update_observations)
            self._logs_observer.update_observations(sample_observations)

        if is_training and step > 0:
            if isinstance(update_observations, dict):
                state = self._agent.update(state, update_key, **update_observations)
            elif isinstance(update_observations, tuple):
                state = self._agent.update(state, update_key, *update_observations)
            else:
                state = self._agent.update(state, update_key, update_observations)

            if self._auto_checkpoint is not None and (step + 1) % self._auto_checkpoint == 0:
                checkpoint_path = os.path.join(self._auto_checkpoint_path, f'rlib-checkpoint-agent-{agent_id}-step-{step + 1}')
                self.save(checkpoint_path, agent_ids=agent_id)

        if isinstance(sample_observations, dict):
            action = self._agent.sample(state, sample_key, **sample_observations)
        elif isinstance(sample_observations, tuple):
            action = self._agent.sample(state, sample_key, *sample_observations)
        else:
            action = self._agent.sample(state, sample_key, sample_observations)

        self._logs_observer.update_agent_state(state)
        self._logs_observer.update_metrics(action, 'action')

        def log_reward(reward: float) -> None:
            self._cumulative_reward += reward
            self._logs_observer.update_metrics(reward, 'reward')
            self._logs_observer.update_metrics(self._cumulative_reward, 'cumulative')

        if 'reward' in all_observations:
            log_reward(all_observations['reward'])
        elif self._ext:
            try:
                if hasattr(self._ext, 'reward'):
                    log_reward(self._ext.reward(**all_observations))
                elif 'reward' in self._ext._observation_functions:
                    log_reward(self._ext._observation_functions['reward'](**all_observations))
            except TypeError:
                pass

        self._agent_containers[agent_id] = AgentContainer(
            state=state,
            key=key,
            action=action,
            step=step + 1
        )

        return action

    def save(self, path: str = None, *, agent_ids: Union[int, list[int]] = None) -> str:
        """
        Saves the state of the experiment to a file in lz4 format. For each agent, both the state and the initialization
        parameters are saved. The extension and loggers settings are saved as well to fully reconstruct the experiment.

        Parameters
        ----------
        path : str, optional
            Path to the checkpoint file. If none specified, saves to the default path.
            If the ``.pkl.lz4`` suffix is not detected, it will be appended automatically.
        agent_ids : int or Array, optional
            The identifier of the agent instance(s) to save. If none specified, saves the state of all agents.
        
        Returns
        -------
        str
            Path to the saved checkpoint file.
        """

        if agent_ids is None:
            agent_ids = list(range(len(self._agent_containers)))
        elif is_scalar(agent_ids):
            agent_ids = [agent_ids]

        agent_containers = [self._agent_containers[agent_id] for agent_id in agent_ids]

        if path is None:
            path = os.path.join(self._default_path, f"rlib-checkpoint-{timestamp()}.pkl.lz4")
        elif path[-8:] != self._lz4_ext:
            path = path + self._lz4_ext

        experiment_state = {
            "agent_type": self._agent_type,
            "agent_params": self._agent_params,
            "agents": {
                agent_id: {
                    "state": agent.state,
                    "key": agent.key,
                    "action": agent.action,
                    "step": agent.step
                } for agent_id, agent in zip(agent_ids, agent_containers)
            },
            "ext_type": self._ext_type,
            "ext_params": self._ext_params,
            "logger_types": self._logger_types,
            "logger_sources": self._logger_sources,
            "logger_params": self._logger_params,
            "auto_checkpoint": self._auto_checkpoint
        }

        with lz4.frame.open(path, 'wb') as f:
            f.write(cloudpickle.dumps(experiment_state))

        return path

    @staticmethod
    def load(
            path: str,
            *,
            agent_params: dict[str, any] = None,
            ext_params: dict[str, any] = None,
            logger_types: Union[type, list[type]] = None,
            logger_sources: Union[Source, list[Source]] = None,
            logger_params: dict[str, any] = None
    ) -> 'RLib':
        """
        Loads the state of the experiment from a file in lz4 format.

        Parameters
        ----------
        path : str
            Path to the checkpoint file.
        agent_params : dict[str, any], optional
            Dictionary of altered agent parameters with their new values, by default None.
        ext_params : dict[str, any], optional
            Dictionary of altered extension parameters with their new values, by default None.
        logger_types : type or list[type], optional
            Types of the selected loggers. Must inherit from the ``BaseLogger`` class.
        logger_sources : Source or list[Source], optional
            Sources to log.
        logger_params : dict, optional
            Parameters of the selected loggers.
        """

        with lz4.frame.open(path, 'rb') as f:
            experiment_state = pickle.loads(f.read())

        rlib = RLib(
            auto_checkpoint=experiment_state["auto_checkpoint"],
            no_ext_mode=experiment_state["ext_type"] is None
        )

        rlib._agent_containers = []

        if experiment_state["ext_type"]:
            if ext_params:
                rlib.set_ext(experiment_state["ext_type"], ext_params)
            else:
                rlib.set_ext(experiment_state["ext_type"], experiment_state["ext_params"])

        if experiment_state["agent_type"]:
            if agent_params:
                rlib.set_agent(experiment_state["agent_type"], agent_params)
            else:
                rlib.set_agent(experiment_state["agent_type"], experiment_state["agent_params"])

        if logger_types:
            rlib.set_loggers(logger_types, logger_sources, logger_params)
        elif experiment_state["logger_types"]:
            rlib.set_loggers(
                experiment_state["logger_types"],
                experiment_state["logger_sources"],
                logger_params if logger_params else experiment_state["logger_params"]
            )

        for agent_id, agent_container in experiment_state["agents"].items():
            while agent_id >= len(rlib._agent_containers):
                rlib.init()

            rlib._agent_containers[agent_id] = AgentContainer(
                state=agent_container["state"],
                key=agent_container["key"],
                action=agent_container["action"],
                step=agent_container["step"]
            )

        return rlib

    def log(self, name: str, value: any) -> None:
        """
        Logs a custom value.

        Parameters
        ----------
        name : str
            The name of the value to log.
        value : any
            The value to log.
        """

        self._logs_observer.update_custom(value, name)

    def to_tflite(self, path: str = None, *, agent_id: int = None, sample_only: bool = False) -> None:
        """
        Converts the agent to a TensorFlow Lite model and saves it to a file.

        Parameters
        ----------
        path : str, optional
            Path to the output file.
        agent_id : int, optional
            The identifier of the agent instance to convert. If specified,
            state of the selected agent will be saved.
        sample_only : bool
            Flag indicating if the method should save only the sample function.
        """

        if not self._agent:
            raise NoAgentError()

        if len(self._agent_containers) == 0:
            self.init()

        if sample_only and agent_id is None:
            raise ValueError("Agent ID must be specified when saving sample function only.")

        if path is None:
            path = self._default_path

        if agent_id is None:
            init_tfl, update_tfl, sample_tfl = self._agent.export(
                init_key=jax.random.key(42)
            )
        else:
            init_tfl, update_tfl, sample_tfl = self._agent.export(
                init_key=self._agent_containers[agent_id].key,
                state=self._agent_containers[agent_id].state,
                sample_only=sample_only
            )

        base_name = self._agent.__class__.__name__
        base_name += f'-{agent_id}-' if agent_id is not None else '-'
        base_name += timestamp()

        with open(os.path.join(path, f'rlib-{base_name}-init.tflite'), 'wb') as f:
            f.write(init_tfl)

        with open(os.path.join(path, f'rlib-{base_name}-sample.tflite'), 'wb') as f:
            f.write(sample_tfl)

        if not sample_only:
            with open(os.path.join(path, f'rlib-{base_name}-update.tflite'), 'wb') as f:
                f.write(update_tfl)



from reinforced_lib.logs.base_logger import BaseLogger, Source, SourceType
from reinforced_lib.logs.csv_logger import CsvLogger
from reinforced_lib.logs.plots_logger import PlotsLogger
from reinforced_lib.logs.stdout_logger import StdoutLogger
from reinforced_lib.logs.tb_logger import TensorboardLogger
from reinforced_lib.logs.wandb_logger import WeightsAndBiasesLogger



from abc import ABC, abstractmethod
from functools import partial
import inspect
from typing import Union

import gymnasium as gym

from reinforced_lib.exts.utils import *
from reinforced_lib.utils.exceptions import IncorrectSpaceError, IncompatibleSpacesError, NoDefaultParameterError


class BaseExt(ABC):
    """
    Container for domain-specific knowledge and functions for a given environment. Provides the transformation
    from the raw observations to the agent update and sample spaces. Stores the default argument values for
    agent initialization.
    """

    def __init__(self) -> None:
        self._observation_functions: dict[str, Callable] = {}
        self._parameter_functions: dict[str, Callable] = {}

        self._add_action_to_observations = False

        for name in dir(self):
            obj = getattr(self, name)

            if hasattr(obj, 'observation_info'):
                self._observation_functions[obj.observation_info.name] = obj

            if hasattr(obj, 'parameter_info'):
                self._parameter_functions[obj.parameter_info.name] = obj

    @property
    @abstractmethod
    def observation_space(self) -> gym.spaces.Space:
        """
        Basic observations of the environment in Gymnasium format.
        """

        pass

    def get_agent_params(
            self,
            agent_type: type = None,
            agent_parameter_space: gym.spaces.Dict = None,
            user_parameters: dict[str, any] = None
    ) -> dict[str, any]:
        """
        Composes agent initialization arguments from values passed by the user and default values stored in the
        parameter functions. Returns a dictionary with the parameters matching the agent parameters space.

        Parameters
        ----------
        agent_type : type, optional
            Type of the selected agent.
        agent_parameter_space : gym.spaces.Dict, optional
            Parameters required by the agents' constructor in Gymnasium format.
        user_parameters : dict, optional
            Parameters provided by the user.

        Returns
        -------
        dict
            Dictionary with the initialization parameters for the agent.
        """

        parameters = user_parameters.copy() if user_parameters else {}

        if agent_parameter_space is None:
            return parameters

        default_parameters = set()

        if agent_type is not None:
            for key, value in inspect.signature(agent_type.__init__).parameters.items():
                if value.default != inspect._empty:
                    default_parameters.add(key)

        for name, space in agent_parameter_space.spaces.items():
            if name in parameters:
                continue

            if name not in self._parameter_functions:
                if name in default_parameters:
                    continue

                raise NoDefaultParameterError(type(self), name, space)

            func = self._parameter_functions[name]
            func_space = func.parameter_info.type

            if space is None or type(space) == type(func_space):
                parameters[name] = func()
            else:
                raise IncompatibleSpacesError(func_space, space)

        return parameters

    def setup_transformations(
            self,
            agent_update_space: gym.spaces.Space = None,
            agent_sample_space: gym.spaces.Space = None
    ) -> None:
        """
        Creates functions that transform raw observations and values provided by the observation functions
        to the agent update and sample spaces.

        Parameters
        ----------
        agent_update_space : gym.spaces.Space, optional
            Observations required by the agent ``update`` function in Gymnasium format.
        agent_sample_space : gym.spaces.Space, optional
            Observations required by the agent ``sample`` function in Gymnasium format.
        """

        if 'action' not in self._observation_functions and \
                isinstance(self.observation_space, gym.spaces.Dict) and \
                'action' not in self.observation_space:

            if 'action' in agent_update_space.spaces:
                self.observation_space['action'] = agent_update_space['action']
            if 'action' in agent_sample_space.spaces:
                self.observation_space['action'] = agent_sample_space['action']

            self._add_action_to_observations = True

        self._update_space_transform = self._transform_spaces(self.observation_space, agent_update_space)
        self._sample_space_transform = self._transform_spaces(self.observation_space, agent_sample_space)

    def _transform_spaces(
            self,
            in_space: gym.spaces.Space,
            out_space: gym.spaces.Space,
            accessor: Union[str, int] = None
    ) -> Callable:
        """
        Creates function that transforms environment observations and values provided by the observation
        functions to a given space. If the ``out_space`` is not defined, returns observations unchanged.

        Parameters
        ----------
        in_space : gym.spaces.Space
            Source space.
        out_space : gym.spaces.Space
            Target space.
        accessor : str or int, optional
            Path to nested observations.

        Returns
        -------
        Callable
            Function that transforms values from ``in_space`` to ``out_space``.
        """

        if out_space is None:
            return lambda *args, **kwargs: None

        simple_types = {
            gym.spaces.Box: test_box,
            gym.spaces.Discrete: test_discrete,
            gym.spaces.MultiBinary: test_multi_binary,
            gym.spaces.MultiDiscrete: test_multi_discrete,
            gym.spaces.Sequence: test_sequence,
            gym.spaces.Space: test_space
        }

        if type(out_space) in simple_types:
            if type(in_space) not in simple_types:
                raise IncompatibleSpacesError(in_space, out_space)

            test_function = simple_types[type(out_space)]

            if test_function(in_space, out_space):
                return partial(self._simple_transform, accessor)

            for observation_function in self._observation_functions.values():
                func_space = observation_function.observation_info.type

                if func_space is None or test_function(func_space, out_space):
                    return observation_function

        if isinstance(out_space, gym.spaces.Dict):
            if not isinstance(in_space, gym.spaces.Dict):
                raise IncompatibleSpacesError(in_space, out_space)

            observations: dict[str, Callable] = {}

            for name, space in out_space.spaces.items():
                if name in in_space.spaces:
                    if type(space) not in simple_types:
                        observations[name] = self._transform_spaces(in_space[name], space, name)
                    elif simple_types[type(space)](in_space[name], space):
                        observations[name] = partial(lambda inner_name, *args, **kwargs: kwargs[inner_name], name)
                    else:
                        raise IncompatibleSpacesError(in_space, space)
                elif name in self._observation_functions:
                    func_space = self._observation_functions[name].observation_info.type

                    if func_space is None or simple_types[type(space)](func_space, space):
                        observations[name] = partial(
                            lambda func, inner_name, inner_accessor, *args, **kwargs:
                            self._function_transform(func, inner_name, inner_accessor, *args, **kwargs),
                            self._observation_functions[name], name, accessor
                        )
                    else:
                        raise IncompatibleSpacesError(func_space, space)
                else:
                    raise IncompatibleSpacesError(in_space, space)

            return partial(self._dict_transform, observations, accessor)

        if isinstance(out_space, gym.spaces.Tuple):
            if not isinstance(in_space, gym.spaces.Tuple) or len(in_space.spaces) != len(out_space.spaces):
                raise IncompatibleSpacesError(in_space, out_space)

            observations: list[Callable] = []

            for i, (agent_space, ext_space) in enumerate(zip(in_space.spaces, out_space.spaces)):
                if type(agent_space) not in simple_types:
                    observations.append(self._transform_spaces(ext_space, agent_space, i))
                elif simple_types[type(agent_space)](ext_space, agent_space):
                    observations.append(partial(lambda inner_i, *args, **kwargs: args[inner_i], i))
                else:
                    raise IncompatibleSpacesError(agent_space, in_space)

            return partial(self._tuple_transform, observations, accessor)

        raise IncorrectSpaceError()

    @staticmethod
    def _get_nested_args(accessor: Union[str, int], *args, **kwargs) -> tuple[tuple, dict]:
        """
        Selects the appropriate nested args or kwargs.

        Parameters
        ----------
        accessor : str or int
            Path to nested observations.
        *args : tuple
            Environment observations.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        tuple[tuple, dict]
            Args and kwargs.
        """

        if accessor is not None:
            if isinstance(accessor, int):
                arguments = args[accessor]
            else:
                arguments = kwargs[accessor]

            if isinstance(arguments, dict):
                return tuple(), arguments
            else:
                return arguments, {}

        return args, kwargs

    def _simple_transform(self, accessor: Union[str, int], *args, **kwargs) -> any:
        """
        Returns the appropriate observation from environment observations.

        Parameters
        ----------
        accessor : str or int
            Path to nested observations.
        *args : tuple
            Environment observations.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        any
            Selected observation from the extension observation space.
        """

        args, kwargs = self._get_nested_args(accessor, *args, **kwargs)

        if len(args) > 0:
            return args[0]
        else:
            first, *_ = kwargs.values()
            return first

    def _function_transform(self, func: Callable, name: str, accessor: Union[str, int], *args, **kwargs) -> any:
        """
        Returns the appropriate observation from the observation function or from environment observations
        if present.

        Parameters
        ----------
        func : Callable
            Function that returns selected observation.
        name : str
            Name of the selected observation.
        accessor : str or int
            Path to nested observations.
        *args : tuple
            Environment observations.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        any
            Selected observation.
        """

        args, kwargs = self._get_nested_args(accessor, *args, **kwargs)

        if name in kwargs:
            return kwargs[name]
        else:
            return func(*args, **kwargs)

    def _dict_transform(self, observations: dict[str, Callable], accessor: Union[str, int], *args, **kwargs) -> dict:
        """
        Returns a dictionary filled with appropriate environment observations and values provided by
        the observation functions.

        Parameters
        ----------
        observations : dict[str, Callable]
            Dictionary with observation names and functions that provide according observations.
        accessor : str or int
            Path to nested observations.
        *args : tuple
            Environment observations.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        dict
            Dictionary with functions providing observations.
        """

        args, kwargs = self._get_nested_args(accessor, *args, **kwargs)
        return {name: func(*args, **kwargs) for name, func in observations.items()}

    def _tuple_transform(self, observations: list[Callable], accessor: Union[str, int], *args, **kwargs) -> tuple:
        """
        Returns a tuple filled with appropriate environment observations and values provided by
        the observation functions.

        Parameters
        ----------
        observations : list[Callable]
            List with functions that provide selected observations.
        accessor : str or int
            Path to nested observations.
        *args : tuple
            Environment observations.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        tuple
            Tuple with functions providing observations.
        """

        args, kwargs = self._get_nested_args(accessor, *args, **kwargs)
        return tuple(func(*args, **kwargs) for func in observations)

    def transform(self, *args, action: any = None, **kwargs) -> tuple[any, any]:
        """
        Transforms raw observations and values provided by the observation functions to the agent observation
        and sample spaces. Provides the last action selected by the agent if it is required by the agent.

        Parameters
        ----------
        *args : tuple
            Environment observations.
        action : any
            The last action selected by the agent.
        **kwargs : dict
            Environment observations.

        Returns
        -------
        tuple[any, any]
            Agent update and sample observations.
        """

        if self._add_action_to_observations:
            kwargs['action'] = action

        return self._update_space_transform(*args, **kwargs), self._sample_space_transform(*args, **kwargs)



from copy import deepcopy
from functools import partial
from typing import Callable

import gymnasium as gym
import jax
import jax.numpy as jnp
import optax
from chex import dataclass, Array, PRNGKey, Scalar, Shape, Numeric
from flax import linen as nn

from reinforced_lib.agents import BaseAgent, AgentState
from reinforced_lib.utils.experience_replay import experience_replay, ExperienceReplay, ReplayBuffer
from reinforced_lib.utils.jax_utils import forward, gradient_step, init


@dataclass
class DDPGState(AgentState):
    r"""
    Container for the state of the deep deterministic policy gradient agent.

    Attributes
    ----------
    q_params : dict
        Parameters of the Q-network.
    q_net_state : dict
        State of the Q-network.
    q_params_target : dict
        Parameters of the target Q-network.
    q_net_state_target : dict
        State of the target Q-network.
    q_opt_state : optax.OptState
        Optimizer state of the Q-network.
    a_params : dict
        Parameters of the policy network.
    a_net_state : dict
        State of the policy network.
    a_params_target : dict
        Parameters of the target policy network.
    a_net_state_target : dict
        State of the target policy network.
    a_opt_state : optax.OptState
        Optimizer state of the policy network.
    replay_buffer : ReplayBuffer
        Experience replay buffer.
    prev_env_state : Array
        Previous environment state.
    noise : Scalar
        Current noise level.
    """

    q_params: dict
    q_net_state: dict
    q_params_target: dict
    q_net_state_target: dict
    q_opt_state: optax.OptState

    a_params: dict
    a_net_state: dict
    a_params_target: dict
    a_net_state_target: dict
    a_opt_state: optax.OptState

    replay_buffer: ReplayBuffer
    prev_env_state: Array
    noise: Scalar


class DDPG(BaseAgent):
    r"""
    Deep deterministic policy gradient [3]_ [4]_ agent with white Gaussian noise exploration and experience replay
    buffer. The agent simultaneously learns a Q-function and a policy. The Q-function is updated using the Bellman
    equation. The policy is learned using the gradient of the Q-function with respect to the policy parameters
    to maximize the Q-value. The agent uses two Q-networks and two policy networks to stabilize the learning process
    and avoid overestimation. The target networks are updated with a soft update. This agent follows the off-policy
    learning paradigm and is suitable for environments with continuous action spaces.

    Parameters
    ----------
    q_network : nn.Module
        Architecture of the Q-networks.
        The input to the network should be two tensors of observations and actions respectively.
    a_network : nn.Module
        Architecture of the policy networks.
    obs_space_shape : Shape
        Shape of the observation space.
    act_space_shape : Shape
        Shape of the action space.
    min_action : Scalar or Array
        Minimum action value.
    max_action : Scalar or Array
        Maximum action value.
    q_optimizer : optax.GradientTransformation, optional
        Optimizer of the Q-networks. If None, the Adam optimizer with learning rate 1e-3 is used.
    a_optimizer : optax.GradientTransformation, optional
        Optimizer of the policy networks. If None, the Adam optimizer with learning rate 1e-3 is used.
    experience_replay_buffer_size : int, default=10000
        Size of the experience replay buffer.
    experience_replay_batch_size : int, default=64
        Batch size of the samples from the experience replay buffer.
    experience_replay_steps : int, default=5
        Number of experience replay steps per update.
    discount : Scalar, default=0.99
        Discount factor. :math:`\gamma = 0.0` means no discount, :math:`\gamma = 1.0` means infinite discount. :math:`0 \leq \gamma \leq 1`
    noise : Scalar, default=(max_action - min_action) / 2
        Initial Gaussian noise level. :math:`0 \leq \sigma`.
    noise_decay : Scalar, default=0.99
        Gaussian noise decay factor. :math:`\sigma_{t+1} = \sigma_{t} * \sigma_{decay}`. :math:`0 \leq \sigma_{decay} \leq 1`.
    noise_min : Scalar, default=0.01
        Minimum Gaussian noise level. :math:`0 \leq \sigma_{min} \leq \sigma`.
    tau : Scalar, default=0.01
        Soft update factor. :math:`\tau = 0.0` means no soft update, :math:`\tau = 1.0` means hard update. :math:`0 \leq \tau \leq 1`.

    References
    ----------
    .. [3] David Silver, Guy Lever, Nicolas Heess, Thomas Degris, Daan Wierstra, and Martin Riedmiller. 2014.
       Deterministic policy gradient algorithms. In Proceedings of the 31st International Conference on International
       Conference on Machine Learning - Volume 32 (ICML'14). JMLR.org, I–387–I–395.
    .. [4] Timothy P. Lillicrap, Jonathan J. Hunt, Alexander Pritzel, Nicolas Heess, Tom Erez, Yuval Tassa, David Silver,
       and Daan Wierstra. 2015. Continuous control with deep reinforcement learning. CoRR abs/1509.02971.
    """

    def __init__(
            self,
            q_network: nn.Module,
            a_network: nn.Module,
            obs_space_shape: Shape,
            act_space_shape: Shape,
            min_action: Numeric,
            max_action: Numeric,
            q_optimizer: optax.GradientTransformation = None,
            a_optimizer: optax.GradientTransformation = None,
            experience_replay_buffer_size: int = 10000,
            experience_replay_batch_size: int = 64,
            experience_replay_steps: int = 5,
            discount: Scalar = 0.99,
            noise: Scalar = None,
            noise_decay: Scalar = 0.99,
            noise_min: Scalar = 0.01,
            tau: Scalar = 0.01
    ) -> None:

        assert experience_replay_buffer_size > experience_replay_batch_size > 0
        assert 0.0 <= discount <= 1.0
        assert 0.0 <= noise_decay <= 1.0
        assert 0.0 <= tau <= 1.0

        if noise is None:
            noise = (max_action - min_action) / 2

        assert 0.0 <= noise
        assert 0.0 <= noise_min <= noise

        if q_optimizer is None:
            q_optimizer = optax.adam(1e-3)
        if a_optimizer is None:
            a_optimizer = optax.adam(1e-3)

        self.obs_space_shape = obs_space_shape if jnp.ndim(obs_space_shape) > 0 else (obs_space_shape,)
        self.act_space_shape = act_space_shape if jnp.ndim(act_space_shape) > 0 else (act_space_shape,)

        er = experience_replay(
            experience_replay_buffer_size,
            experience_replay_batch_size,
            self.obs_space_shape,
            self.act_space_shape
        )

        self.init = jax.jit(partial(
            self.init,
            obs_space_shape=self.obs_space_shape,
            act_space_shape=self.act_space_shape,
            q_network=q_network,
            a_network=a_network,
            q_optimizer=q_optimizer,
            a_optimizer=a_optimizer,
            er=er,
            noise=noise
        ))
        self.update = jax.jit(partial(
            self.update,
            q_step_fn=partial(
                gradient_step,
                optimizer=q_optimizer,
                loss_fn=partial(self.q_loss_fn, q_network=q_network, a_network=a_network, discount=discount)
            ),
            a_step_fn=partial(
                gradient_step,
                optimizer=a_optimizer,
                loss_fn=partial(self.a_loss_fn, q_network=q_network, a_network=a_network)
            ),
            er=er,
            experience_replay_steps=experience_replay_steps,
            noise_decay=noise_decay,
            noise_min=noise_min,
            tau=tau
        ))
        self.sample = jax.jit(partial(
            self.sample,
            a_network=a_network,
            min_action=min_action,
            max_action=max_action
        ))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'obs_space_shape': gym.spaces.Sequence(gym.spaces.Box(1, jnp.inf, (1,), int)),
            'act_space_shape': gym.spaces.Sequence(gym.spaces.Box(1, jnp.inf, (1,), int)),
            'min_action': gym.spaces.Sequence(gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)),
            'max_action': gym.spaces.Sequence(gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)),
            'experience_replay_buffer_size': gym.spaces.Box(1, jnp.inf, (1,), int),
            'experience_replay_batch_size': gym.spaces.Box(1, jnp.inf, (1,), int),
            'discount': gym.spaces.Box(0.0, 1.0, (1,), float),
            'noise': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'noise_decay': gym.spaces.Box(0.0, 1.0, (1,), float),
            'noise_min': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'tau': gym.spaces.Box(0.0, 1.0, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'env_state': gym.spaces.Box(-jnp.inf, jnp.inf, self.obs_space_shape, float),
            'action': gym.spaces.Box(-jnp.inf, jnp.inf, self.act_space_shape, float),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float),
            'terminal': gym.spaces.MultiBinary(1)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'env_state': gym.spaces.Box(-jnp.inf, jnp.inf, self.obs_space_shape, float)
        })

    @property
    def action_space(self) -> gym.spaces.Box:
        return gym.spaces.Box(-jnp.inf, jnp.inf, self.act_space_shape, float)

    @staticmethod
    def init(
            key: PRNGKey,
            obs_space_shape: Shape,
            act_space_shape: Shape,
            q_network: nn.Module,
            a_network: nn.Module,
            q_optimizer: optax.GradientTransformation,
            a_optimizer: optax.GradientTransformation,
            er: ExperienceReplay,
            noise: Scalar
    ) -> DDPGState:
        r"""
        Initializes the Q-networks and the policy networks, optimizers, and experience replay buffer.
        The first state of the environment is assumed to be a tensor of zeros.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        obs_space_shape : Shape
            The shape of the observation space.
        act_space_shape : Shape
            The shape of the action space.
        q_network : nn.Module
            The Q-network.
        a_network : nn.Module
            The policy network.
        q_optimizer : optax.GradientTransformation
            The Q-network optimizer.
        a_optimizer : optax.GradientTransformation
            The policy network optimizer.
        er : ExperienceReplay
            The experience replay buffer.
        noise : Scalar
            The initial noise value.

        Returns
        -------
        DDPGState
            Initial state of the deep deterministic policy gradient agent.
        """

        s_dummy = jnp.empty(obs_space_shape)
        a_dummy = jnp.empty(act_space_shape)

        key, q_key, a_key = jax.random.split(key, 3)

        q_params, q_net_state = init(q_network, q_key, s_dummy, a_dummy)
        a_params, a_net_state = init(a_network, a_key, s_dummy)

        q_opt_state = q_optimizer.init(q_params)
        a_opt_state = a_optimizer.init(a_params)
        replay_buffer = er.init()

        return DDPGState(
            q_params=q_params,
            q_net_state=q_net_state,
            q_params_target=deepcopy(q_params),
            q_net_state_target=deepcopy(q_net_state),
            q_opt_state=q_opt_state,
            a_params=a_params,
            a_net_state=a_net_state,
            a_params_target=deepcopy(a_params),
            a_net_state_target=deepcopy(a_net_state),
            a_opt_state=a_opt_state,
            replay_buffer=replay_buffer,
            prev_env_state=jnp.zeros(obs_space_shape),
            noise=noise
        )

    @staticmethod
    def q_loss_fn(
            q_params: dict,
            key: PRNGKey,
            state: DDPGState,
            batch: tuple,
            q_network: nn.Module,
            a_network: nn.Module,
            discount: Scalar
    ) -> tuple[Scalar, dict]:
        r"""
        Loss is the mean squared Bellman error :math:`\mathcal{L}(\theta) = \mathbb{E}_{s, a, r, s'} \left[ \left( r
        + \gamma \max Q'(s', \pi'(s')) - Q(s, a) \right)^2 \right]` where :math:`s` is the current state, :math:`a`
        is the current action, :math:`r` is the reward, :math:`s'` is the next state, :math:`\gamma` is  the discount
        factor, :math:`Q(s, a)` is the Q-value of the main Q-network, :math:`Q'(s, a)` is the Q-value of the target
        Q-network, and :math:`\pi'(s)` is the action of the target policy network. The policy network parameters
        are considered as fixed. Loss can be calculated on a batch of transitions.

        Parameters
        ----------
        q_params : dict
            The parameters of the Q-network.
        key : PRNGKey
            A PRNG key used as the random key.
        state : DDPGState
            The state of the deep deterministic policy gradient agent.
        batch : tuple
            A batch of transitions from the experience replay buffer.
        q_network : nn.Module
            The Q-network.
        a_network : nn.Module
            The policy network.
        discount : Scalar
            The discount factor.

        Returns
        -------
        tuple[Scalar, dict]
            The loss and the new state of the Q-network.
        """

        states, actions, rewards, terminals, next_states = batch
        q_key, q_target_key, a_target_key = jax.random.split(key, 3)

        q_values, q_net_state = forward(q_network, q_params, state.q_net_state, q_key, states, actions)

        actions_target, _ = forward(a_network, state.a_params_target, state.a_net_state_target, a_target_key, next_states)
        q_values_target, _ = forward(q_network, state.q_params_target, state.q_net_state_target, q_target_key, next_states, actions_target)
        target = rewards + (1 - terminals) * discount * q_values_target

        target = jax.lax.stop_gradient(target)
        loss = optax.l2_loss(q_values, target).mean()

        return loss, q_net_state

    @staticmethod
    def a_loss_fn(
            a_params: dict,
            key: PRNGKey,
            state: DDPGState,
            batch: tuple,
            q_network: nn.Module,
            a_network: nn.Module
    ) -> tuple[Scalar, dict]:
        r"""
        The policy network is updated using the gradient of the Q-network to maximize the Q-value of the current state
        and action :math:`\max_{\theta} \mathbb{E}_{s, a} \left[ Q(s, \pi_{\theta}(s)) \right]`. Q-network parameters are
        considered as fixed. The policy network can be updated on a batch of transitions.

        Parameters
        ----------
        a_params : dict
            The parameters of the policy network.
        key : PRNGKey
            A PRNG key used as the random key.
        state : DDPGState
            The state of the deep deterministic policy gradient agent.
        batch : tuple
            A batch of transitions from the experience replay buffer.
        q_network : nn.Module
            The Q-network.
        a_network : nn.Module
            The policy network.

        Returns
        -------
        tuple[Scalar, dict]
            The loss and the new state of the policy network.
        """

        states, _, _, _, _ = batch
        a_key, q_key = jax.random.split(key)

        actions, a_net_state = forward(a_network, a_params, state.a_net_state, a_key, states)
        q_values, _ = forward(q_network, state.q_params, state.q_net_state, q_key, states, actions)
        loss = -jnp.mean(q_values)

        return loss, a_net_state

    @staticmethod
    def update(
            state: DDPGState,
            key: PRNGKey,
            env_state: Array,
            action: Array,
            reward: Scalar,
            terminal: bool,
            q_step_fn: Callable,
            a_step_fn: Callable,
            er: ExperienceReplay,
            experience_replay_steps: int,
            noise_decay: Scalar,
            noise_min: Scalar,
            tau: Scalar
    ) -> DDPGState:
        r"""
        Appends the transition to the experience replay buffer and performs ``experience_replay_steps`` steps.
        Each step consists of sampling a batch of transitions from the experience replay buffer, calculating the
        Q-network loss and the policy network loss using ``q_loss_fn`` and ``a_loss_fn`` respectively, performing
        a gradient step on both networks, and soft updating the target networks.  Soft update of the parameters
        is defined as :math:`\theta_{target} = \tau \theta + (1 - \tau) \theta_{target}`.The noise parameter is
        decayed by ``noise_decay``.

        Parameters
        ----------
        state : DDPGState
            The current state of the double Q-learning agent.
        key : PRNGKey
            A PRNG key used as the random key.
        env_state : Array
            The current state of the environment.
        action : Array
            The action taken by the agent.
        reward : Scalar
            The reward received by the agent.
        terminal : bool
            Whether the episode has terminated.
        q_step_fn : Callable
            The function that performs a single gradient step on the Q-network.
        a_step_fn : Callable
            The function that performs a single gradient step on the policy network.
        er : ExperienceReplay
            The experience replay buffer.
        experience_replay_steps : int
            The number of experience replay steps.
        noise_decay : Scalar
            The decay rate of the noise parameter.
        noise_min : Scalar
            The minimum value of the noise parameter.
        tau : Scalar
            The soft update parameter.

        Returns
        -------
        DDPGState
            The updated state of the deep deterministic policy gradient agent.
        """

        replay_buffer = er.append(state.replay_buffer, state.prev_env_state, action, reward, terminal, env_state)

        def for_loop_fn(_: int, carry: tuple) -> tuple:
            q_params, q_net_state, q_params_target, q_net_state_target, q_opt_state, \
            a_params, a_net_state, a_params_target, a_net_state_target, a_opt_state, key = carry

            batch_key, q_network_key, a_network_key, key = jax.random.split(key, 4)
            batch = er.sample(replay_buffer, batch_key)

            q_params, q_net_state, q_opt_state, _ = q_step_fn(q_params, (q_network_key, state, batch), q_opt_state)
            a_params, a_net_state, a_opt_state, _ = a_step_fn(a_params, (a_network_key, state, batch), a_opt_state)

            q_params_target, q_net_state_target = optax.incremental_update((q_params, q_net_state), (q_params_target, q_net_state_target), tau)
            a_params_target, a_net_state_target = optax.incremental_update((a_params, a_net_state), (a_params_target, a_net_state_target), tau)

            return q_params, q_net_state, q_params_target, q_net_state_target, q_opt_state, \
                   a_params, a_net_state, a_params_target, a_net_state_target, a_opt_state, key

        q_params, q_net_state, q_params_target, q_net_state_target, q_opt_state, \
        a_params, a_net_state, a_params_target, a_net_state_target, a_opt_state, _ = jax.lax.fori_loop(
            0, experience_replay_steps, for_loop_fn,
            (state.q_params, state.q_net_state, state.q_params_target, state.q_net_state_target, state.q_opt_state,
            state.a_params, state.a_net_state, state.a_params_target, state.a_net_state_target, state.a_opt_state, key)
        )

        return DDPGState(
            q_params=q_params,
            q_net_state=q_net_state,
            q_params_target=q_params_target,
            q_net_state_target=q_net_state_target,
            q_opt_state=q_opt_state,
            a_params=a_params,
            a_net_state=a_net_state,
            a_params_target=a_params_target,
            a_net_state_target=a_net_state_target,
            a_opt_state=a_opt_state,
            replay_buffer=replay_buffer,
            prev_env_state=env_state,
            noise=jnp.maximum(state.noise * noise_decay, noise_min)
        )

    @staticmethod
    def sample(
            state: DDPGState,
            key: PRNGKey,
            env_state: Array,
            a_network: nn.Module,
            min_action: Scalar,
            max_action: Scalar
    ) -> Numeric:
        r"""
        Calculates deterministic action using the policy network. Then adds white Gaussian noise with standard
        deviation ``state.noise`` to the action and clips it to the range :math:`[min\_action, max\_action]`.

        Parameters
        ----------
        state : DDPGState
            The state of the double Q-learning agent.
        key : PRNGKey
            A PRNG key used as the random key.
        env_state : Array
            The current state of the environment.
        a_network : nn.Module
            The policy network.
        min_action : Scalar or Array
            The minimum value of the action.
        max_action : Scalar or Array
            The maximum value of the action.

        Returns
        -------
        Scalar or Array
            Selected action.
        """

        network_key, noise_key = jax.random.split(key)

        action, _ = forward(a_network, state.a_params, state.a_net_state, network_key, env_state)
        action += jax.random.normal(noise_key, action.shape) * state.noise

        return jnp.clip(action, min_action, max_action)



import gymnasium as gym
import jax.numpy as jnp

from reinforced_lib.exts import BaseExt, observation, parameter


class Gymnasium(BaseExt):
    """
    Gymnasium [1]_ extension. Simplifies interaction of RL agents with the Gymnasium environments by providing
    the environment state, reward, terminal flag, and shapes of the observation and action spaces.

    Parameters
    ----------
    env_id : str
        Name of the Gymnasium environment.

    References
    ----------
    .. [1] Gymnasium https://gymnasium.farama.org
    """

    def __init__(self, env_id: str) -> None:
        self.env = gym.make(env_id)
        super().__init__()

    observation_space = gym.spaces.Dict({})

    @observation()
    def env_state(self, env_state, reward, terminal, truncated, info, *args, **kwargs) -> any:
        return env_state

    @observation(observation_type=gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float))
    def reward(self, env_state, reward, terminal, truncated, info, *args, **kwargs) -> float:
        return reward

    @observation(observation_type=gym.spaces.MultiBinary(1))
    def terminal(self, env_state, reward, terminal, truncated, info, *args, **kwargs) -> bool:
        return terminal or truncated

    @parameter(parameter_type=gym.spaces.Sequence(gym.spaces.Box(1, jnp.inf, (1,), int)))
    def obs_space_shape(self) -> tuple:
        return self.env.observation_space.shape

    @parameter(parameter_type=gym.spaces.Sequence(gym.spaces.Box(1, jnp.inf, (1,), int)))
    def act_space_shape(self) -> tuple:
        return self.env.action_space.shape

    @parameter(parameter_type=gym.spaces.Box(1, jnp.inf, (1,), int))
    def act_space_size(self) -> int:
        if isinstance(self.env.action_space, gym.spaces.Discrete):
            return self.env.action_space.n

        raise AttributeError()

    @parameter(parameter_type=gym.spaces.Sequence(gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)))
    def min_action(self) -> tuple:
        return self.env.action_space.low

    @parameter(parameter_type=gym.spaces.Sequence(gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)))
    def max_action(self) -> tuple:
        return self.env.action_space.high



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, PRNGKey, Scalar

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class ThompsonSamplingState(AgentState):
    """
    Container for the state of the Thompson sampling agent.

    Attributes
    ----------
    alpha : Array
        Number of successful tries for each arm.
    beta : Array
        Number of failed tries for each arm.
    """

    alpha: Array
    beta: Array


class ThompsonSampling(BaseAgent):
    r"""
    Contextual Bernoulli Thompson sampling agent with the exponential smoothing. The implementation is inspired by  the
    work of Krotov et al. [7]_. Thompson sampling is based on a beta distribution with parameters related to the number
    of successful and failed attempts. Higher values of the parameters decrease the entropy of the distribution while
    changing the ratio of the parameters shifts the expected value.

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}`.
    decay : float, default=0.0
        Decay rate. If equal to zero, smoothing is not applied. :math:`w \geq 0`.

    References
    ----------
    .. [7] Alexander Krotov, Anton Kiryanov and Evgeny Khorov. 2020. Rate Control With Spatial Reuse
       for Wi-Fi 6 Dense Deployments. IEEE Access. 8. 168898-168909.
    """

    def __init__(self, n_arms: int, decay: Scalar = 0.0) -> None:
        assert decay >= 0

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=self.n_arms))
        self.update = jax.jit(partial(self.update, decay=decay))
        self.sample = jax.jit(self.sample)

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'decay': gym.spaces.Box(0.0, jnp.inf, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'n_successful': gym.spaces.Box(0, jnp.inf, (1,), int),
            'n_failed': gym.spaces.Box(0, jnp.inf, (1,), int),
            'delta_time': gym.spaces.Box(0.0, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'context': gym.spaces.Box(-jnp.inf, jnp.inf, (self.n_arms,), float)
        })

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(key: PRNGKey, n_arms: int) -> ThompsonSamplingState:
        r"""
        Creates and initializes an instance of the Thompson sampling agent for ``n_arms`` arms. The :math:`\mathbf{\alpha}`
        and :math:`\mathbf{\beta}` vectors are set to zero to create a non-informative prior distribution.
        The ``last_decay`` is also set to zero.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.

        Returns
        -------
        ThompsonSamplingState
            Initial state of the Thompson sampling agent.
        """

        return ThompsonSamplingState(
            alpha=jnp.zeros((n_arms, 1)),
            beta=jnp.zeros((n_arms, 1))
        )

    @staticmethod
    def update(
            state: ThompsonSamplingState,
            key: PRNGKey,
            action: int,
            n_successful: int,
            n_failed: int,
            delta_time: Scalar,
            decay: Scalar
    ) -> ThompsonSamplingState:
        r"""
        Thompson sampling can be adjusted to non-stationary environments by exponential smoothing of values of
        vectors :math:`\mathbf{\alpha}` and :math:`\mathbf{\beta}` which increases the entropy of a distribution
        over time. Given a result of trial :math:`s`, we apply the following equations for each action :math:`a`:

        .. math::
           \begin{gather}
              \mathbf{\alpha}_{t + 1}(a) = \mathbf{\alpha}_t(a) e^{\frac{-\Delta t}{w}} + \mathbb{1}_{A = a} \cdot s , \\
              \mathbf{\beta}_{t + 1}(a) = \mathbf{\beta}_t(a) e^{\frac{-\Delta t}{w}} + \mathbb{1}_{A = a} \cdot (1 - s) ,
           \end{gather}

        where :math:`\Delta t` is the time elapsed since the last action selection and :math:`w` is the decay rate.

        Parameters
        ----------
        state : ThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        n_successful : int
            Number of successful tries.
        n_failed : int
            Number of failed tries.
        delta_time : float
            Time elapsed since the last action selection.
        decay : float
            Decay rate.

        Returns
        -------
        ThompsonSamplingState
            Updated agent state.
        """

        smoothing_value = jnp.exp(-decay * delta_time)

        return ThompsonSamplingState(
            alpha=(state.alpha * smoothing_value).at[action].add(n_successful),
            beta=(state.beta * smoothing_value).at[action].add(n_failed)
        )

    @staticmethod
    def sample(
            state: ThompsonSamplingState,
            key: PRNGKey,
            context: Array
    ) -> int:
        r"""
        The Thompson sampling policy is stochastic. The algorithm draws :math:`q_a` from the distribution
        :math:`\operatorname{Beta}(1 + \mathbf{\alpha}(a), 1 + \mathbf{\beta}(a))` for each arm :math:`a`.
        The next action is selected as

        .. math::
           A = \operatorname*{argmax}_{a \in \mathscr{A}} q_a r_a ,

        where :math:`r_a` is contextual information for the arm :math:`a`, and :math:`\mathscr{A}` is a set
        of all actions.

        Parameters
        ----------
        state : ThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        context : Array
            One-dimensional array of features for each arm.

        Returns
        -------
        int
            Selected action.
        """

        success_prob = jax.random.beta(key, 1 + state.alpha, 1 + state.beta).flatten()
        action = jnp.argmax(success_prob * context)

        return action



import gymnasium as gym
import jax.numpy as jnp

from reinforced_lib.exts import BaseExt, observation, parameter


class BasicMab(BaseExt):
    """
    Basic multi-armed bandit (MAB) extension for Reinforced-lib. This extension can be used with MAB algorithms
    which do not require any additional information about the environment apart from the number of arms.
    """

    def __init__(self, n_arms: int) -> None:
        super().__init__()
        self.n = n_arms

    observation_space = gym.spaces.Dict({})

    @parameter(parameter_type=gym.spaces.Box(1, jnp.inf, (1,), int))
    def n_arms(self) -> int:
        return self.n

    @observation(observation_type=gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float))
    def reward(self, reward, *args, **kwargs) -> float:
        return reward



from typing import Callable, NamedTuple

import gymnasium as gym
import jax.numpy as jnp


class ObservationInfo(NamedTuple):
    """
    Description of the observation function that provides one of the values from the agent observation space.

    Attributes
    ----------
    name : str
        Name of the provided observation.
    type : gym.spaces.Space
        Type of the provided value in Gymnasium format.
    """

    name: str
    type: gym.spaces.Space


class ParameterInfo(NamedTuple):
    """
    Description of the parameter function that provides one of the parameters of the agent constructor.

    Attributes
    ----------
    name : str
        Name of the provided parameter.
    type : gym.spaces.Space
        Type of the provided parameter in Gymnasium format.
    """

    name: str
    type: gym.spaces.Space


def observation(observation_name: str = None, observation_type: gym.spaces.Space = None) -> Callable:
    """
    Decorator used to annotate the observation functions.

    Parameters
    ----------
    observation_name : str, optional
        Name of the provided observation.
    observation_type : gym.spaces.Space, optional
        Type of the provided value in Gymnasium format.

    Returns
    -------
    Callable
        Function that returns the appropriate observation.
    """

    def decorator(function):
        name = observation_name if observation_name is not None else function.__name__
        function.observation_info = ObservationInfo(name, observation_type)
        return function

    return decorator


def parameter(parameter_name: str = None, parameter_type: gym.spaces.Space = None) -> Callable:
    """
    Decorator used to annotate the parameter functions.

    Parameters
    ----------
    parameter_name : str, optional
        Name of the provided parameter.
    parameter_type : gym.spaces.Space, optional
        Type of the provided parameter in Gymnasium format.

    Returns
    -------
    Callable
        Function that returns the appropriate parameter.
    """

    def decorator(function):
        name = parameter_name if parameter_name is not None else function.__name__
        function.parameter_info = ParameterInfo(name, parameter_type)
        return function

    return decorator


def test_box(a: gym.spaces.Space, b: gym.spaces.Box) -> bool:
    """
    Tests if the space ``a`` is identical to the gym.space.Box space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.Box
        Box space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    return isinstance(a, gym.spaces.Box) and \
        jnp.array_equal(a.low, b.low) and \
        jnp.array_equal(a.high, b.high) and \
        a.shape == b.shape and \
        a.dtype == b.dtype


def test_discrete(a: gym.spaces.Space, b: gym.spaces.Discrete) -> bool:
    """
    Tests if the space ``a`` is identical to the gym.space.Discrete space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.Discrete
        Discrete space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    return isinstance(a, gym.spaces.Discrete) and \
        a.n == b.n and \
        a.start == b.start


def test_multi_binary(a: gym.spaces.Space, b: gym.spaces.MultiBinary) -> bool:
    """
    Tests if the space ``a`` is identical to the gym.space.MultiBinary space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.MultiBinary
        MultiBinary space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    return isinstance(a, gym.spaces.MultiBinary) and \
        jnp.array_equal(a.n, b.n)


def test_multi_discrete(a: gym.spaces.Space, b: gym.spaces.MultiDiscrete) -> bool:
    """
    Tests if the space ``a`` is identical to the gym.space.MultiDiscrete space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.MultiDiscrete
        MultiDiscrete space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    return isinstance(a, gym.spaces.MultiDiscrete) and \
        jnp.array_equal(a.nvec, b.nvec) and \
        a.dtype == b.dtype


def test_sequence(a: gym.spaces.Space, b: gym.spaces.Sequence) -> bool:
    """
    Tests if the space ``a`` is identical to the gym.space.Sequence space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.Sequence
        Sequence space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    if not isinstance(a, gym.spaces.Sequence):
        return False

    if isinstance(b, gym.spaces.Box):
        return test_box(a, b)
    elif isinstance(b, gym.spaces.Discrete):
        return test_discrete(a, b)
    elif isinstance(b, gym.spaces.MultiBinary):
        return test_multi_binary(a, b)
    elif isinstance(b, gym.spaces.MultiDiscrete):
        return test_multi_discrete(a, b)
    else:
        return test_space(a, b)


def test_space(a: gym.spaces.Space, b: gym.spaces.Space) -> bool:
    """
    Tests if the space ``a`` is identical to the space ``b``.

    Parameters
    ----------
    a : gym.spaces.Space
        Space ``a``.
    b : gym.spaces.Space
        Space ``b``.

    Returns
    -------
    bool
        Result of the comparison.
    """

    return a.shape == b.shape and \
        a.dtype == b.dtype



import json
import os.path
from collections import defaultdict
from copy import deepcopy

import jax.numpy as jnp
import numpy as np
from chex import Array, Scalar

from reinforced_lib.logs import BaseLogger, Source
from reinforced_lib.utils import timestamp


class CsvLogger(BaseLogger):
    """
    Logger that saves values in CSV format. It saves the logged values to the CSV file when the experiment is finished.
    ``CsvLogger`` synchronizes the logged values in time. It means that if the same source is logged twice in a row,
    the step number will be incremented for all columns and the logger will move to the next row.

    Parameters
    ----------
    csv_path : str, default="~/rlib-logs-[date]-[time].csv"
        Path to the output file.
    """

    def __init__(self, csv_path: str = None, **kwargs) -> None:
        super().__init__(**kwargs)

        if csv_path is None:
            csv_path = f'rlib-logs-{timestamp()}.csv'
            csv_path = os.path.join(os.path.expanduser("~"), csv_path)

        self._csv_path = csv_path
        self._current_values = set()
        self._step = 0

        self._values = defaultdict(list)
        self._steps = defaultdict(list)

    def finish(self) -> None:
        """
        Saves the logged values to the CSV file.
        """

        file = open(self._csv_path, 'w')
        file.write(','.join(self._values.keys()) + '\n')

        rows, cols = self._step + 1, len(self._values)
        csv_array = np.full((rows, cols), fill_value='', dtype=object)

        for j, (name, values) in enumerate(self._values.items()):
            for i, v in enumerate(values):
                csv_array[self._steps[name][i], j] = v

        for row in csv_array:
            file.write(','.join(map(str, row)) + '\n')

        file.close()

    def log_scalar(self, source: Source, value: Scalar, *_) -> None:
        """
        Logs a scalar as a standard value in a column.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        """

        self._log(source, value)

    def log_array(self, source: Source, value: Array, *_) -> None:
        """
        Logs an array as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        """

        if isinstance(value, (np.ndarray, jnp.ndarray)):
            value = value.tolist()

        self._log(source, f"\"{json.dumps(value)}\"")

    def log_dict(self, source: Source, value: dict, *_) -> None:
        """
        Logs a dictionary as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : dict
            Dictionary to log.
        """

        self._log(source, f"\"{json.dumps(value)}\"")

    def log_other(self, source: Source, value: any, *_) -> None:
        """
        Logs an object as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        """

        self._log(source, f"\"{json.dumps(value)}\"")

    def _log(self, source: Source, value: any) -> None:
        """
        Saves the logged value and controls the current step.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value to log.
        """

        name = self.source_to_name(source)

        if name in self._current_values:
            self._step += 1
            self._current_values.clear()

        self._current_values.add(name)
        self._values[name].append(deepcopy(value))
        self._steps[name].append(self._step)



import json
from collections import defaultdict

from chex import Array, Scalar
from tensorboardX import SummaryWriter

from reinforced_lib.logs import BaseLogger, Source


class TensorboardLogger(BaseLogger):
    """
    Logger that saves values in TensorBoard [2]_ format. Offers a possibility to log to Comet [3]_.
    ``TensorboardLogger`` synchronizes the logged values in time. This means that if the same source
    is logged less often than other sources, the step will be increased accordingly to maintain the
    appropriate spacing between the values on the x-axis.

    Parameters
    ----------
    tb_log_dir : str, optional
        Path to the output directory. If None, the default directory is used.
    tb_comet_config : dict, optional
        Configuration for the Comet logger. If None, the logger is disabled.
    tb_sync_steps : bool, default=False
        Set to ``True`` if you want to synchronize the logged values in time.

    References
    ----------
    .. [2] TensorBoard. https://www.tensorflow.org/tensorboard
    .. [3] Comet. https://www.comet.ml
    """

    def __init__(
            self,
            tb_log_dir: str = None,
            tb_comet_config: dict[str, any] = None,
            tb_sync_steps: bool = False,
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if tb_comet_config is None:
            tb_comet_config = {'disabled': True}

        self._sync_steps = tb_sync_steps
        self._current_values = set()
        self._step = 0

        self._writer = SummaryWriter(log_dir=tb_log_dir, comet_config=tb_comet_config)
        self._steps = defaultdict(int)

    def finish(self) -> None:
        """
        Closes the summary writer.
        """

        self._writer.close()

    def log_scalar(self, source: Source, value: Scalar, *_) -> None:
        """
        Adds a given scalar to the summary writer.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        """

        name = self.source_to_name(source)
        step = self._get_step(name)
        self._writer.add_scalar(name, value, step)

    def log_array(self, source: Source, value: Array, *_) -> None:
        """
        Log values from an array to the same plot. Creates multiple line plots for each value in the array.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        """

        name = self.source_to_name(source)
        step = self._get_step(name)
        self._writer.add_scalars(name, {str(i): float(v) for i, v in enumerate(value)}, step)

    def log_dict(self, source: Source, value: dict, *_) -> None:
        """
        Logs a dictionary as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : dict
            Dictionary to log.
        """

        self.log_other(source, value, None)

    def log_other(self, source: Source, value: any, *_) -> None:
        """
        Logs an object as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        """

        name = self.source_to_name(source)
        step = self._get_step(name)
        self._writer.add_text(name, json.dumps(value), step)

    def _get_step(self, name: str) -> int:
        """
        Returns the current step for a given source.

        Parameters
        ----------
        name : str
            Name of the source.

        Returns
        -------
        int
            Current step for the given source.
        """

        if self._sync_steps:
            if name in self._current_values:
                self._step += 1
                self._current_values.clear()

            self._current_values.add(name)
            step = self._step
        else:
            step = self._steps[name] + 1

        self._steps[name] = step
        return step



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, PRNGKey, Scalar

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class NormalThompsonSamplingState(AgentState):
    """
    Container for the state of the normal Thompson sampling agent.

    Attributes
    ----------
    alpha : Array
        The concentration parameter of the inverse-gamma distribution.
    beta : Array
        The scale parameter of the inverse-gamma distribution.
    lam : Array
        The number of observations.
    mu : Array
        The mean of the normal distribution.
    """

    alpha: Array
    beta: Array
    lam: Array
    mu: Array


class NormalThompsonSampling(BaseAgent):
    r"""
    Normal Thompson sampling agent [10]_. The normal-inverse-gamma distribution is a conjugate prior for the normal
    distribution with unknown mean and variance. The parameters of the distribution are updated after each observation.
    The mean of the normal distribution is sampled from the normal-inverse-gamma distribution and the action with
    the highest expected value is selected.

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}` .
    alpha : float
        See also ``NormalThompsonSamplingState`` for interpretation. :math:`\alpha > 0`.
    beta : float
        See also ``NormalThompsonSamplingState`` for interpretation. :math:`\beta > 0`.
    lam : float
        See also ``NormalThompsonSamplingState`` for interpretation. :math:`\lambda > 0`.
    mu : float
        See also ``NormalThompsonSamplingState`` for interpretation. :math:`\mu \in \mathbb{R}`.

    References
    ----------
    .. [10] Kevin P. Murphy. 2007. Conjugate Bayesian analysis of the Gaussian distribution.
       https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf
    """

    def __init__(
            self,
            n_arms: int,
            alpha: Scalar,
            beta: Scalar,
            lam: Scalar,
            mu: Scalar
    ) -> None:
        assert alpha > 0
        assert beta > 0
        assert lam >= 0

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=self.n_arms, alpha=alpha, beta=beta, lam=lam, mu=mu))
        self.update = jax.jit(self.update)
        self.sample = jax.jit(self.sample)

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'alpha': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'beta': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'lam': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'mu': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Space:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(
            key: PRNGKey,
            n_arms: int,
            alpha: Scalar,
            beta: Scalar,
            lam: Scalar,
            mu: Scalar
    ) -> NormalThompsonSamplingState:
        r"""
        Creates and initializes an instance of the normal Thompson sampling agent for ``n_arms`` arms and the
        given initial parameters for the prior distribution.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.
        alpha : float
            See also ``NormalThompsonSamplingState`` for interpretation.
        beta : float
            See also ``NormalThompsonSamplingState`` for interpretation.
        lam : float
            See also ``NormalThompsonSamplingState`` for interpretation.
        mu : float
            See also ``NormalThompsonSamplingState`` for interpretation.
        Returns
        -------
        NormalThompsonSamplingState
            Initial state of the normal Thompson sampling agent.
        """

        return NormalThompsonSamplingState(
            alpha=jnp.full((n_arms, 1), alpha),
            beta=jnp.full((n_arms, 1), beta),
            lam=jnp.full((n_arms, 1), lam),
            mu=jnp.full((n_arms, 1), mu)
        )

    @staticmethod
    def update(
            state: NormalThompsonSamplingState,
            key: PRNGKey,
            action: int,
            reward: Scalar
    ) -> NormalThompsonSamplingState:
        r"""
        Normal Thompson sampling update according to [10]_.

        .. math::
          \begin{align}
            \alpha_{t + 1}(a) &= \alpha_t(a) + \frac{1}{2} \\
            \beta_{t + 1}(a) &= \beta_t(a) + \frac{\lambda_t(a) (r_t(a) - \mu_t(a))^2}{2 (\lambda_t(a) + 1)} \\
            \lambda_{t + 1}(a) &= \lambda_t(a) + 1 \\
            \mu_{t + 1}(a) &= \frac{\mu_t(a) \lambda_t(a) + r_t(a)}{\lambda_t(a) + 1}
          \end{align}

        Parameters
        ----------
        state : NormalThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : Float
            Reward obtained upon execution of action.

        Returns
        -------
        NormalThompsonSamplingState
            Updated agent state.
        """

        lam = state.lam[action]
        mu = state.mu[action]

        return NormalThompsonSamplingState(
            alpha=state.alpha.at[action].add(1 / 2),
            beta=state.beta.at[action].add((lam * jnp.square(reward - mu)) / (2 * (lam + 1))),
            lam=state.lam.at[action].add(1),
            mu=state.mu.at[action].set((mu * lam + reward) / (lam + 1))
        )

    @staticmethod
    def inverse_gamma(key: PRNGKey, concentration: Array, scale: Array) -> Array:
        r"""
        Samples from the inverse gamma distribution. Implementation is based on the gamma distribution and the
        following dependence:

        .. math::
          \begin{gather}
              X \sim \operatorname{Gamma}(\alpha, \beta) \\
              \frac{1}{X} \sim \operatorname{Inverse-gamma}(\alpha, \frac{1}{\beta})
          \end{gather}

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        concentration : Array
            The concentration parameter of the inverse-gamma distribution.
        scale : Array
            The scale parameter of the inverse-gamma distribution.

        Returns
        -------
        Array
            Sampled values from the inverse gamma distribution.
        """

        gamma = jax.random.gamma(key, concentration) / scale
        return 1 / gamma

    @staticmethod
    def sample(state: NormalThompsonSamplingState, key: PRNGKey) -> int:
        r"""
        The normal Thompson sampling policy is stochastic. The algorithm draws :math:`q_a` from the distribution
        :math:`\operatorname{Normal}(\mu(a), \operatorname{scale}(a)/\sqrt{\lambda(a)})` for each arm :math:`a` where
        :math:`\text{scale}(a)` is sampled from the inverse gamma distribution with parameters :math:`\alpha(a)` and
        :math:`\beta(a)`. The next action is selected as :math:`A = \operatorname*{argmax}_{a \in \mathscr{A}} q_a`,
        where :math:`\mathscr{A}` is a set of all actions.

        Parameters
        ----------
        state : NormalThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.

        Returns
        -------
        int
            Selected action.
        """

        loc_key, scale_key = jax.random.split(key)

        scale = jnp.sqrt(NormalThompsonSampling.inverse_gamma(scale_key, state.alpha, state.beta))
        loc = state.mu + jax.random.normal(loc_key, shape=state.mu.shape) * scale / jnp.sqrt(state.lam)

        return jnp.argmax(loc)



from reinforced_lib.agents.deep.ddpg import DDPG
from reinforced_lib.agents.deep.ddqn import DDQN
from reinforced_lib.agents.deep.expected_sarsa import ExpectedSarsa
from reinforced_lib.agents.deep.dqn import DQN



from copy import deepcopy
from functools import partial
from typing import Callable

import gymnasium as gym
import jax
import jax.numpy as jnp
import optax
from chex import dataclass, Array, PRNGKey, Scalar, Shape
from flax import linen as nn

from reinforced_lib.agents import BaseAgent, AgentState
from reinforced_lib.utils.experience_replay import experience_replay, ExperienceReplay, ReplayBuffer
from reinforced_lib.utils.jax_utils import forward, gradient_step, init


@dataclass
class DQNState(AgentState):
    r"""
    Container for the state of the deep Q-learning agent.

    Attributes
    ----------
    params : dict
        Parameters of the Q-network.
    net_state : dict
        State of the Q-network.
    opt_state : optax.OptState
        Optimizer state.
    replay_buffer : ReplayBuffer
        Experience replay buffer.
    prev_env_state : Array
        Previous environment state.
    epsilon : Scalar
        :math:`\epsilon`-greedy parameter.
    """

    params: dict
    net_state: dict
    opt_state: optax.OptState

    replay_buffer: ReplayBuffer
    prev_env_state: Array
    epsilon: Scalar


class DQN(BaseAgent):
    r"""
    Deep Q-learning agent [1]_ with :math:`\epsilon`-greedy exploration and experience replay buffer. The agent uses
    a deep neural network to approximate the Q-value function. The Q-network is trained to minimize the Bellman
    error. This agent follows the off-policy learning paradigm and is suitable for environments with discrete action
    spaces.

    Parameters
    ----------
    q_network : nn.Module
        Architecture of the Q-network.
    obs_space_shape : Shape
        Shape of the observation space.
    act_space_size : int
        Size of the action space.
    optimizer : optax.GradientTransformation, optional
        Optimizer of the Q-network. If None, the Adam optimizer with learning rate 1e-3 is used.
    experience_replay_buffer_size : int, default=10000
        Size of the experience replay buffer.
    experience_replay_batch_size : int, default=64
        Batch size of the samples from the experience replay buffer.
    experience_replay_steps : int, default=5
        Number of experience replay steps per update.
    discount : Scalar, default=0.99
        Discount factor. :math:`\gamma = 0.0` means no discount, :math:`\gamma = 1.0` means infinite discount. :math:`0 \leq \gamma \leq 1`
    epsilon : Scalar, default=1.0
        Initial :math:`\epsilon`-greedy parameter. :math:`0 \leq \epsilon \leq 1`.
    epsilon_decay : Scalar, default=0.999
        Epsilon decay factor. :math:`\epsilon_{t+1} = \epsilon_{t} * \epsilon_{decay}`. :math:`0 \leq \epsilon_{decay} \leq 1`.
    epsilon_min : Scalar, default=0.01
        Minimum :math:`\epsilon`-greedy parameter. :math:`0 \leq \epsilon_{min} \leq \epsilon`.

    References
    ----------
    .. [1] Mnih, V., Kavukcuoglu, K., Silver, D., Graves, A., Antonoglou, I., Wierstra, D. & Riedmiller, M. (2013).
       Playing Atari with Deep Reinforcement Learning.
    """

    def __init__(
            self,
            q_network: nn.Module,
            obs_space_shape: Shape,
            act_space_size: int,
            optimizer: optax.GradientTransformation = None,
            experience_replay_buffer_size: int = 10000,
            experience_replay_batch_size: int = 64,
            experience_replay_steps: int = 5,
            discount: Scalar = 0.99,
            epsilon: Scalar = 1.0,
            epsilon_decay: Scalar = 0.999,
            epsilon_min: Scalar = 0.001
    ) -> None:

        assert experience_replay_buffer_size > experience_replay_batch_size > 0
        assert 0.0 <= discount <= 1.0
        assert 0.0 <= epsilon <= 1.0
        assert 0.0 <= epsilon_decay <= 1.0

        if optimizer is None:
            optimizer = optax.adam(1e-3)

        self.obs_space_shape = obs_space_shape if jnp.ndim(obs_space_shape) > 0 else (obs_space_shape,)
        self.act_space_size = act_space_size

        er = experience_replay(
            experience_replay_buffer_size,
            experience_replay_batch_size,
            self.obs_space_shape,
            (1,)
        )

        self.init = jax.jit(partial(
            self.init,
            obs_space_shape=self.obs_space_shape,
            q_network=q_network,
            optimizer=optimizer,
            er=er,
            epsilon=epsilon
        ))
        self.update = jax.jit(partial(
            self.update,
            step_fn=partial(
                gradient_step,
                optimizer=optimizer,
                loss_fn=partial(self.loss_fn, q_network=q_network, discount=discount)
            ),
            er=er,
            experience_replay_steps=experience_replay_steps,
            epsilon_decay=epsilon_decay,
            epsilon_min=epsilon_min
        ))
        self.sample = jax.jit(partial(
            self.sample,
            q_network=q_network,
            act_space_size=act_space_size
        ))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'obs_space_shape': gym.spaces.Sequence(gym.spaces.Box(1, jnp.inf, (1,), int)),
            'act_space_size': gym.spaces.Box(1, jnp.inf, (1,), int),
            'experience_replay_buffer_size': gym.spaces.Box(1, jnp.inf, (1,), int),
            'experience_replay_batch_size': gym.spaces.Box(1, jnp.inf, (1,), int),
            'discount': gym.spaces.Box(0.0, 1.0, (1,), float),
            'epsilon': gym.spaces.Box(0.0, 1.0, (1,), float),
            'epsilon_decay': gym.spaces.Box(0.0, 1.0, (1,), float),
            'epsilon_min': gym.spaces.Box(0.0, 1.0, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'env_state': gym.spaces.Box(-jnp.inf, jnp.inf, self.obs_space_shape, float),
            'action': gym.spaces.Discrete(self.act_space_size),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float),
            'terminal': gym.spaces.MultiBinary(1)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'env_state': gym.spaces.Box(-jnp.inf, jnp.inf, self.obs_space_shape, float)
        })

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(self.act_space_size)

    @staticmethod
    def init(
            key: PRNGKey,
            obs_space_shape: Shape,
            q_network: nn.Module,
            optimizer: optax.GradientTransformation,
            er: ExperienceReplay,
            epsilon: Scalar
    ) -> DQNState:
        r"""
        Initializes the Q-network, optimizer and experience replay buffer with given parameters.
        The first state of the environment is assumed to be a tensor of zeros.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        obs_space_shape : Shape
            The shape of the observation space.
        q_network : nn.Module
            The Q-network.
        optimizer : optax.GradientTransformation
            The optimizer.
        er : ExperienceReplay
            The experience replay buffer.
        epsilon : Scalar
            The initial :math:`\epsilon`-greedy parameter.

        Returns
        -------
        DQNState
            Initial state of the deep Q-learning agent.
        """

        x_dummy = jnp.empty(obs_space_shape)
        params, net_state = init(q_network, key, x_dummy)

        opt_state = optimizer.init(params)
        replay_buffer = er.init()

        return DQNState(
            params=params,
            net_state=net_state,
            opt_state=opt_state,
            replay_buffer=replay_buffer,
            prev_env_state=jnp.zeros(obs_space_shape),
            epsilon=epsilon
        )

    @staticmethod
    def loss_fn(
            params: dict,
            key: PRNGKey,
            net_state: dict,
            params_target: dict,
            net_state_target: dict,
            batch: tuple,
            q_network: nn.Module,
            discount: Scalar
    ) -> tuple[Scalar, dict]:
        r"""
        Loss is the mean squared Bellman error :math:`\mathcal{L}(\theta) = \mathbb{E}_{s, a, r, s'} \left[ \left( r +
        \gamma \max_{a'} Q(s', a') - Q(s, a) \right)^2 \right]` where :math:`s` is the current state, :math:`a` is the
        current action, :math:`r` is the reward, :math:`s'` is the next state, :math:`\gamma` is  the discount factor, 
        :math:`Q(s, a)` is the Q-value of the state-action pair. Loss can be calculated on a batch of transitions.

        Parameters
        ----------
        params : dict
            The parameters of the Q-network.
        key : PRNGKey
            A PRNG key used as the random key.
        net_state : dict
            The state of the Q-network.
        params_target : dict
            The parameters of the target Q-network.
        net_state_target : dict
            The state of the target Q-network.
        batch : tuple
            A batch of transitions from the experience replay buffer.
        q_network : nn.Module
            The Q-network.
        discount : Scalar
            The discount factor.

        Returns
        -------
        Tuple[Scalar, dict]
            The loss and the new state of the Q-network.
        """

        states, actions, rewards, terminals, next_states = batch
        q_key, q_target_key = jax.random.split(key)

        q_values, net_state = forward(q_network, params, net_state, q_key, states)
        q_values = jnp.take_along_axis(q_values, actions.astype(int), axis=-1)

        q_values_target, _ = forward(q_network, params_target, net_state_target, q_target_key, next_states)
        target = rewards + (1 - terminals) * discount * jnp.max(q_values_target, axis=-1, keepdims=True)

        target = jax.lax.stop_gradient(target)
        loss = optax.l2_loss(q_values, target).mean()

        return loss, net_state

    @staticmethod
    def update(
            state: DQNState,
            key: PRNGKey,
            env_state: Array,
            action: Array,
            reward: Scalar,
            terminal: bool,
            step_fn: Callable,
            er: ExperienceReplay,
            experience_replay_steps: int,
            epsilon_decay: Scalar,
            epsilon_min: Scalar
    ) -> DQNState:
        r"""
        Appends the transition to the experience replay buffer and performs ``experience_replay_steps`` steps.
        Each step consists of sampling a batch of transitions from the experience replay buffer, calculating the loss
        using the ``loss_fn`` function and performing a gradient step on the Q-network. The :math:`\epsilon`-greedy
        parameter is decayed by ``epsilon_decay``.

        Parameters
        ----------
        state : DQNState
            The current state of the deep Q-learning agent.
        key : PRNGKey
            A PRNG key used as the random key.
        env_state : Array
            The current state of the environment.
        action : Array
            The action taken by the agent.
        reward : Scalar
            The reward received by the agent.
        terminal : bool
            Whether the episode has terminated.
        step_fn : Callable
            The function that performs a single gradient step on the Q-network.
        er : ExperienceReplay
            The experience replay buffer.
        experience_replay_steps : int
            The number of experience replay steps.
        epsilon_decay : Scalar
            The decay rate of the :math:`\epsilon`-greedy parameter.
        epsilon_min : Scalar
            The minimum value of the :math:`\epsilon`-greedy parameter.

        Returns
        -------
        DQNState
            The updated state of the deep Q-learning agent.
        """

        replay_buffer = er.append(state.replay_buffer, state.prev_env_state, action, reward, terminal, env_state)
        params_target, net_state_target = deepcopy(state.params), deepcopy(state.net_state)

        def for_loop_fn(_: int, carry: tuple) -> tuple:
            params, net_state, opt_state, key = carry
            batch_key, network_key, key = jax.random.split(key, 3)

            loss_params = (network_key, net_state, params_target, net_state_target, er.sample(replay_buffer, batch_key))
            params, net_state, opt_state, _ = step_fn(params, loss_params, opt_state)

            return params, net_state, opt_state, key

        params, net_state, opt_state, _ = jax.lax.fori_loop(
            0, experience_replay_steps * er.is_ready(replay_buffer), for_loop_fn,
            (state.params, state.net_state, state.opt_state, key)
        )

        return DQNState(
            params=params,
            net_state=net_state,
            opt_state=opt_state,
            replay_buffer=replay_buffer,
            prev_env_state=env_state,
            epsilon=jax.lax.max(state.epsilon * epsilon_decay, epsilon_min)
        )

    @staticmethod
    def sample(
            state: DQNState,
            key: PRNGKey,
            env_state: Array,
            q_network: nn.Module,
            act_space_size: int
    ) -> int:
        r"""
        Samples random action with probability :math:`\epsilon` and the greedy action with probability
        :math:`1 - \epsilon`. The greedy action is the action with the highest Q-value.
        
        Parameters
        ----------
        state : DQNState
            The state of the deep Q-learning agent.
        key : PRNGKey
            A PRNG key used as the random key.
        env_state : Array
            The current state of the environment.
        q_network : nn.Module
            The Q-network.
        act_space_size : int
            The size of the action space.

        Returns
        -------
        int
            Selected action.
        """

        network_key, action_key = jax.random.split(key)

        q, _ = forward(q_network, state.params, state.net_state, network_key, env_state)
        max_q = (q == q.max()).astype(float)
        probs = (1 - state.epsilon) * max_q / jnp.sum(max_q) + state.epsilon / q.shape[0]

        return jax.random.choice(action_key, act_space_size, p=probs.flatten())



import unittest
import os
import jax
import jax.numpy as jnp

from reinforced_lib.agents.mab import UCB
from reinforced_lib.rlib import RLib
from reinforced_lib.logs import *


class TestRLibSerialization(unittest.TestCase):

    checkpoint_path = os.path.join(os.path.expanduser("~"), "checkpoint.pkl.lz4")
    arms_probs = jnp.array([1.0, 1.0, 0.99, 0.97, 0.91, 0.77, 0.32, 0.05, 0.01, 0.0, 0.0, 0.0])
    time = jnp.linspace(0, 10, 1000)
    t_change = jnp.max(time) / 2
    key = jax.random.key(42)


    def run_experiment(self, reload: bool, new_decay: float = None) -> list[int]:
        rl = RLib(
            agent_type=UCB,
            agent_params={'n_arms': len(self.arms_probs), 'c': 0.1},
            no_ext_mode=True,
            logger_types=CsvLogger,
            logger_sources=['n_failed', 'n_successful', ('action', SourceType.METRIC)],
            logger_params={'csv_path': f'output_reload={reload}_new-decay={new_decay}.csv'}
        )

        actions = []
        a = 0

        reloaded = not reload
        for t in self.time:
            r = int(jax.random.uniform(self.key) < self.arms_probs[a])
            observations = {
                'action': a,
                'reward': r
            }

            a = rl.sample(update_observations=observations)
            actions.append(a)

            if t > self.t_change and not reloaded:
                save_path = rl.save()

                if new_decay:
                    rl = RLib.load(save_path, agent_params={'n_arms': len(self.arms_probs), 'c': 0.0})
                else:
                    rl = RLib.load(save_path)
                reloaded = True
        
        return actions
    

    def test_reload(self):
        """
        Tests if the experiment state is fully reconstructable after reload.
        """

        actions_straight = self.run_experiment(reload=False)
        actions_reload = self.run_experiment(reload=True)
        self.assertTrue(jnp.array_equal(actions_straight, actions_reload))
    

    def test_params_alter(self):
        """
        Tests if the experiment alters after the parameter alters.
        """

        actions_straight = self.run_experiment(reload=False)
        actions_reload = self.run_experiment(reload=True, new_decay=True)
        self.assertFalse(jnp.array_equal(actions_straight, actions_reload))


if __name__ == "__main__":
    unittest.main()



import jax
import jax.numpy as jnp
from chex import PRNGKey, Scalar

from reinforced_lib.agents.mab.normal_thompson_sampling import NormalThompsonSampling, NormalThompsonSamplingState


class LogNormalThompsonSampling(NormalThompsonSampling):
    r"""
    Log-normal Thompson sampling agent. This algorithm is designed to handle positive rewards by transforming
    them into the log-space. For more details, refer to the documentation on ``NormalThompsonSampling``.
    """

    @staticmethod
    def update(
            state: NormalThompsonSamplingState,
            key: PRNGKey,
            action: int,
            reward: Scalar
    ) -> NormalThompsonSamplingState:
        r"""
        Log-normal Thompson sampling update. The update is analogous to the one in ``NormalThompsonSampling`` except
        that the reward is transformed into the log-space.

        Parameters
        ----------
        state : NormalThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : Float
            Reward obtained upon execution of action.

        Returns
        -------
        NormalThompsonSamplingState
            Updated agent state.
        """

        return NormalThompsonSampling.update(state, key, action, jnp.log(reward))

    @staticmethod
    def sample(state: NormalThompsonSamplingState, key: PRNGKey) -> int:
        r"""
        Sampling actions is analogous to the one in ``NormalThompsonSampling`` except that the expected value of
        the log-normal distribution is computed instead of the expected value of the normal distribution.

        Parameters
        ----------
        state : NormalThompsonSamplingState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.

        Returns
        -------
        int
            Selected action.
        """

        loc_key, scale_key = jax.random.split(key)

        scale = jnp.sqrt(NormalThompsonSampling.inverse_gamma(scale_key, state.alpha, state.beta))
        loc = state.mu + jax.random.normal(loc_key, shape=state.mu.shape) * scale / jnp.sqrt(state.lam)
        log_normal_mean = jnp.exp(loc + 0.5 * jnp.square(scale))

        return jnp.argmax(log_normal_mean)



from functools import partial

import jax
import jax.numpy as jnp
from chex import Array, PRNGKey, dataclass

from reinforced_lib.agents import AgentState, BaseAgent


@dataclass
class MaskedState(AgentState):
    agent_state: AgentState
    mask: Array


class Masked(BaseAgent):
    r"""
    Meta agent supporting dynamic change of number of arms.

    **This agent is highly experimental and is expected to be used with an extreme caution.**
    In particular, this agent makes the following strong assumptions:

    - Each entry in the base agent state has the first dimension corresponding to an arm.
    - The base agent must be stochastic as this agent uses rejection sampling to choose a possible action

    Example usage of the agent can be found in the test `test/experimental/test_masked.py`.

    Parameters
    ----------
    agent : BaseAgent
        A MAB agent type which actions are masked.
    mask : Array
        Binary mask array of the length equal to the number of arms. Positive entries are the masked actions.
    """

    def __init__(self, agent: BaseAgent, mask: Array) -> None:
        self.init = jax.jit(partial(self.init, agent=agent, mask=mask))
        self.update = jax.jit(partial(self.update, agent=agent))
        self.sample = jax.jit(partial(self.sample, agent=agent))

    @staticmethod
    def init(key: PRNGKey, *args, agent: BaseAgent, mask: Array, **kwargs) -> MaskedState:
        return MaskedState(agent_state=agent.init(key, *args, **kwargs), mask=mask)

    @staticmethod
    def update(state: MaskedState, key: PRNGKey, *args, agent: BaseAgent, **kwargs) -> MaskedState:
        tree_mask = jax.tree.map(lambda _: jnp.expand_dims(state.mask, 1), state.agent_state)
        agent_state = agent.update(state.agent_state, key, *args, **kwargs)
        agent_state = jax.tree.map(lambda s, ns, m: jnp.where(m, s, ns), state.agent_state, agent_state, tree_mask)
        return MaskedState(agent_state=agent_state, mask=state.mask)

    @staticmethod
    def sample(state: MaskedState, key: PRNGKey, *args, agent: BaseAgent, **kwargs) -> int:
        sample_key, while_key = jax.random.split(key, 2)
        action = agent.sample(state.agent_state, sample_key, *args, **kwargs)

        def cond_fn(carry: tuple) -> bool:
            action, _ = carry
            return state.mask[action]

        def body_fn(carry: tuple) -> tuple:
            action, key = carry
            sample_key, key = jax.random.split(key)
            key = jax.random.fold_in(key, action)
            action = agent.sample(state.agent_state, sample_key, *args, **kwargs)
            return action, key

        action, _ = jax.lax.while_loop(cond_fn, body_fn, (action, while_key))
        return action



import json
from copy import deepcopy

import jax.numpy as jnp
import numpy as np
from chex import Array, Scalar

from reinforced_lib.logs import BaseLogger, Source


class StdoutLogger(BaseLogger):
    """
    Logger that writes values to the standard output.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._values = {}

    def finish(self) -> None:
        """
        Prints the last row if there are any unprinted values left.
        """

        if len(self._values) > 0:
            print('\t'.join(f'{n}: {v}' for n, v in self._values.items()))

    def log_scalar(self, source: Source, value: Scalar, custom: bool) -> None:
        """
        Logs a scalar as the standard value.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        custom : bool
            Whether the source is a custom source.
        """

        self._log(source, value, custom)

    def log_array(self, source: Source, value: Array, custom: bool) -> None:
        """
        Logs an array as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        custom : bool
            Whether the source is a custom source.
        """

        if isinstance(value, (np.ndarray, jnp.ndarray)):
            value = value.tolist()

        self._log(source, json.dumps(value), custom)

    def log_dict(self, source: Source, value: dict, custom: bool) -> None:
        """
        Logs a dictionary as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : dict
            Dictionary to log.
        custom : bool
            Whether the source is a custom source.
        """

        self._log(source, json.dumps(value), custom)

    def log_other(self, source: Source, value: any, custom: bool) -> None:
        """
        Logs an object as a JSON string.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        custom : bool
            Whether the source is a custom source.
        """

        self._log(source, json.dumps(value), custom)

    def _log(self, source: Source, value: any, custom: bool) -> None:
        """
        Prints a new row to the standard output if there is a new value for a
        standard source or the source is custom.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        custom : bool
            Whether the source is a custom source.
        """

        name = self.source_to_name(source)

        if not custom:
            if name in self._values:
                print('\t'.join(f'{n}: {v}' for n, v in self._values.items()))
                self._values = {}

            self._values[name] = deepcopy(value)
        else:
            print(f'{name}: {value}')



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, Scalar, PRNGKey

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class UCBState(AgentState):
    """
    Container for the state of the UCB agent.

    Attributes
    ----------
    R : Array
        Sum of the rewards obtained for each arm.
    N : Array
        Number of tries for each arm.
    """

    R: Array
    N: Array


class UCB(BaseAgent):
    r"""
    UCB agent with optional discounting. The main idea behind this algorithm is to introduce a preference factor
    in its policy, so that the selection of the next action is based on both the current estimation of the
    action-value function and the uncertainty of this estimation.

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}`.
    c : float
        Degree of exploration. :math:`c \geq 0`.
    gamma : float, default=1.0
        If less than one, a discounted UCB algorithm [8]_ is used. :math:`\gamma \in (0, 1]`.

    References
    ----------
    .. [8] Aurélien Garivier, Eric Moulines. 2008. On Upper-Confidence Bound Policies for Non-Stationary
       Bandit Problems. 10.48550/ARXIV.0805.3415.
    """

    def __init__(
            self,
            n_arms: int,
            c: Scalar,
            gamma: Scalar = 1.0
    ) -> None:
        assert c >= 0
        assert 0 < gamma <= 1

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=n_arms))
        self.update = jax.jit(partial(self.update, gamma=gamma))
        self.sample = jax.jit(partial(self.sample, c=c))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'c': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'gamma': gym.spaces.Box(0.0, 1.0, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Space:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(
            key: PRNGKey,
            n_arms: int
    ) -> UCBState:
        """
        Creates and initializes instance of the UCB agent for ``n_arms`` arms.  The sum of the rewards is set to zero
        and the number of tries is set to one for each arm.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.

        Returns
        -------
        UCBState
            Initial state of the UCB agent.
        """

        return UCBState(
            R=jnp.zeros((n_arms, 1)),
            N=jnp.zeros((n_arms, 1))
        )

    @staticmethod
    def update(
        state: UCBState,
        key: PRNGKey,
        action: int,
        reward: Scalar,
        gamma: Scalar
    ) -> UCBState:
        r"""
        In the stationary case, the sum of the rewards for a given arm is increased by reward :math:`r` obtained after
        step :math:`t` and the number of tries for the corresponding arm is incremented. In the non-stationary case,
        the update follows the equations

        .. math::
           \begin{gather}
              R_{t + 1}(a) = \mathbb{1}_{A_t = a} r + \gamma R_t(a) , \\
              N_{t + 1}(a) = \mathbb{1}_{A_t = a} + \gamma N_t(a).
           \end{gather}

        Parameters
        ----------
        state : UCBState
            Current state of agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : float
            Reward collected by the agent after taking the previous action.
        gamma : float
            Discount factor.

        Returns
        -------
        UCBState
            Updated agent state.
        """

        return UCBState(
            R=(gamma * state.R).at[action].add(reward),
            N=(gamma * state.N).at[action].add(1)
        )

    @staticmethod
    def sample(
        state: UCBState,
        key: PRNGKey,
        c: Scalar
    ) -> int:
        r"""
        UCB agent follows the policy

        .. math::
           A = \operatorname*{argmax}_{a \in \mathscr{A}} \left[ Q(a) + c \sqrt{\frac{\ln \left( {\sum_{a' \in \mathscr{A}}} N(a') \right) }{N(a)}} \right] .

        where :math:`\mathscr{A}` is a set of all actions and :math:`Q` is calculated as :math:`Q(a) = \frac{R(a)}{N(a)}`.
        The second component of the sum represents a sort of upper bound on the value of :math:`Q`, where :math:`c`
        behaves like a confidence interval and the square root - like an approximation of the :math:`Q` function
        estimation uncertainty. Note that the UCB policy is deterministic (apart from choosing between several optimal actions).

        Parameters
        ----------
        state : UCBState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        c : float
            Degree of exploration.

        Returns
        -------
        int
            Selected action.
        """

        Q = jnp.where(state.N == 0., jnp.inf, state.R / state.N)
        t = jnp.maximum(1, jnp.sum(state.N))

        ucb = Q + c * jnp.sqrt(jnp.log(t) / jnp.maximum(state.N, 1))
        max_ucb = (ucb == jnp.max(ucb)).astype(float)
        probs = max_ucb / jnp.sum(max_ucb)

        return jax.random.choice(key, probs.shape[0], p=probs.flatten())



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, Scalar, PRNGKey

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class SoftmaxState(AgentState):
    r"""
    Container for the state of the Softmax agent.

    Attributes
    ----------
    H : Array
        Preference for each arm.
    r : float
        Average of all obtained rewards :math:`\bar{R}`.
    n : int
        Step number.
    """

    H: Array
    r: Scalar
    n: int


class Softmax(BaseAgent):
    r"""
    Softmax agent with baseline and optional exponential recency-weighted average update. It learns a preference
    function :math:`H`, which indicates a preference of selecting one arm over others. Algorithm policy can be
    controlled by the temperature parameter :math:`\tau`. The implementation is inspired by the work of Sutton
    and Barto [5]_. **Note:** For this agent, some environments find it very beneficial to use 64-bit JAX mode!

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}`.
    lr : float
        Step size. :math:`lr > 0`.
    alpha : float, default=0.0
        If non-zero, exponential recency-weighted average is used to update :math:`\bar{R}`. :math:`\alpha \in [0, 1]`.
    tau : float, default=1.0
        Temperature parameter. :math:`\tau > 0`.
    multiplier : float, default=1.0
        Multiplier for the reward. :math:`multiplier > 0`.
    """

    def __init__(
            self,
            n_arms: int,
            lr: Scalar,
            alpha: Scalar = 0.0,
            tau: Scalar = 1.0,
            multiplier: Scalar = 1.0
    ) -> None:
        assert lr > 0
        assert 0 <= alpha <= 1
        assert tau > 0
        assert multiplier > 0

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=n_arms))
        self.update = jax.jit(partial(self.update, lr=lr, alpha=alpha, tau=tau, multiplier=multiplier))
        self.sample = jax.jit(partial(self.sample, tau=tau))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'lr': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'alpha': gym.spaces.Box(0.0, 1.0, (1,), float),
            'tau': gym.spaces.Box(0.0, jnp.inf, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(
            key: PRNGKey,
            n_arms: int
    ) -> SoftmaxState:
        r"""
        Creates and initializes instance of the Softmax agent for ``n_arms`` arms. Preferences :math:`H` for each arm
        are set to zero, as well as the average of all rewards :math:`\bar{R}`. The step number :math:`n` is
        initialized to one.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.

        Returns
        -------
        SoftmaxState
            Initial state of the Softmax agent.
        """

        return SoftmaxState(
            H=jnp.zeros((n_arms, 1)),
            r=0.0,
            n=1
        )

    @staticmethod
    def update(
        state: SoftmaxState,
        key: PRNGKey,
        action: int,
        reward: Scalar,
        lr: Scalar,
        alpha: Scalar,
        tau: Scalar,
        multiplier: Scalar
    ) -> SoftmaxState:
        r"""
        Preferences :math:`H` can be learned by stochastic gradient ascent. The softmax algorithm searches
        for such a set of preferences that maximizes the expected reward :math:`\mathbb{E}[R]`.
        The updates of :math:`H` for each action :math:`a` are calculated as:

        .. math::
           H_{t + 1}(a) = H_t(a) + \alpha (R_t - \bar{R}_t)(\mathbb{1}_{A_t = a} - \pi_t(a)),

        where :math:`\bar{R_t}` is the average of all rewards up to but not including step :math:`t`
        (by definition :math:`\bar{R}_1 = R_1`). The derivation of given formula can be found in [5]_.

        In the stationary case, :math:`\bar{R_t}` can be calculated as
        :math:`\bar{R}_{t + 1} = \bar{R}_t + \frac{1}{t} \lbrack R_t - \bar{R}_t \rbrack`. To improve the
        algorithm's performance in the non-stationary case, we apply
        :math:`\bar{R}_{t + 1} = \bar{R}_t + \alpha \lbrack R_t - \bar{R}_t \rbrack` with a constant
        step size :math:`\alpha`.

        Reward :math:`R_t` is multiplied by ``multiplier`` before updating preferences to allow for
        more flexible reward scaling while keeping the algorithm's properties.

        Parameters
        ----------
        state : SoftmaxState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : float
            Reward collected by the agent after taking the previous action.
        lr : float
            Step size.
        alpha : float
            Exponential recency-weighted average factor (used when :math:`\alpha > 0`).
        tau : float
            Temperature parameter.
        multiplier : float
            Multiplier for the reward.

        Returns
        -------
        SoftmaxState
            Updated agent state.
        """

        reward *= multiplier

        r = jnp.where(state.n == 1, reward, state.r)
        pi = jax.nn.softmax(state.H / tau)

        return SoftmaxState(
            H=state.H + lr * (reward - r) * (jnp.zeros_like(state.H).at[action].set(1) - pi),
            r=r + (reward - r) * jnp.where(alpha == 0, 1 / state.n, alpha),
            n=state.n + 1
        )

    @staticmethod
    def sample(
        state: SoftmaxState,
        key: PRNGKey,
        tau: Scalar
    ) -> int:
        r"""
        The policy of the Softmax algorithm is stochastic. The algorithm draws the next action from the softmax
        distribution. The probability of selecting action :math:`i` is calculated as:

        .. math::
           softmax(H)_i = \frac{\exp(H_i / \tau)}{\sum_{h \in H} \exp(h / \tau)} .

        Parameters
        ----------
        state : SoftmaxState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        tau : float
            Temperature parameter.

        Returns
        -------
        int
            Selected action.
        """

        return jax.random.categorical(key, state.H.flatten() / tau)



from reinforced_lib.agents.mab.e_greedy import EGreedy
from reinforced_lib.agents.mab.exp3 import Exp3
from reinforced_lib.agents.mab.normal_thompson_sampling import NormalThompsonSampling
from reinforced_lib.agents.mab.lognormal_thompson_sampling import LogNormalThompsonSampling
from reinforced_lib.agents.mab.softmax import Softmax
from reinforced_lib.agents.mab.thompson_sampling import ThompsonSampling
from reinforced_lib.agents.mab.ucb import UCB
